from moto.stepfunctions.parser.asl.component.intrinsic.argument.function_argument import (
    FunctionArgument,
)


class FunctionArgumentString(FunctionArgument):
    _value: str

    def __init__(self, string: str):
        super().__init__(value=string)



import json
from typing import Optional

from moto.core.exceptions import JsonRESTError


class IoTClientError(JsonRESTError):
    code = 400


class ResourceNotFoundException(IoTClientError):
    def __init__(self, msg: Optional[str] = None):
        self.code = 404
        super().__init__(
            "ResourceNotFoundException", msg or "The specified resource does not exist"
        )


class InvalidRequestException(IoTClientError):
    def __init__(self, msg: Optional[str] = None):
        self.code = 400
        super().__init__("InvalidRequestException", msg or "The request is not valid.")


class InvalidStateTransitionException(IoTClientError):
    def __init__(self, msg: Optional[str] = None):
        self.code = 409
        super().__init__(
            "InvalidStateTransitionException",
            msg or "An attempt was made to change to an invalid state.",
        )


class VersionConflictException(IoTClientError):
    def __init__(self, name: str):
        self.code = 409
        super().__init__(
            "VersionConflictException",
            f"The version for thing {name} does not match the expected version.",
        )


class CertificateStateException(IoTClientError):
    def __init__(self, msg: str, cert_id: str):
        self.code = 406
        super().__init__("CertificateStateException", f"{msg} Id: {cert_id}")


class DeleteConflictException(IoTClientError):
    def __init__(self, msg: str):
        self.code = 409
        super().__init__("DeleteConflictException", msg)


class ResourceAlreadyExistsException(IoTClientError):
    def __init__(self, msg: str, resource_id: str, resource_arn: str):
        self.code = 409
        super().__init__(
            "ResourceAlreadyExistsException", msg or "The resource already exists."
        )
        self.description = json.dumps(
            {
                "message": self.message,
                "resourceId": resource_id,
                "resourceArn": resource_arn,
            }
        )


class VersionsLimitExceededException(IoTClientError):
    def __init__(self, name: str):
        self.code = 409
        super().__init__(
            "VersionsLimitExceededException",
            f"The policy {name} already has the maximum number of versions (5)",
        )


class ThingStillAttached(IoTClientError):
    def __init__(self, name: str):
        self.code = 409
        super().__init__(
            "InvalidRequestException",
            f"Cannot delete. Thing {name} is still attached to one or more principals",
        )



from typing import Any

from moto.core.exceptions import RESTError

EXCEPTION_RESPONSE = """<?xml version="1.0"?>
<ErrorResponse xmlns="http://elasticache.amazonaws.com/doc/2015-02-02/">
  <Error>
    <Type>Sender</Type>
    <Code>{{ error_type }}</Code>
    <Message>{{ message }}</Message>
  </Error>
  <{{ request_id_tag }}>30c0dedb-92b1-4e2b-9be4-1188e3ed86ab</{{ request_id_tag }}>
</ErrorResponse>"""


class ElasticBeanstalkException(RESTError):
    code = 400
    extended_templates = {"ecerror": EXCEPTION_RESPONSE}
    env = RESTError.extended_environment(extended_templates)

    def __init__(self, code: str, message: str, **kwargs: Any):
        kwargs.setdefault("template", "ecerror")
        super().__init__(code, message)


class InvalidParameterValueError(RESTError):
    def __init__(self, message: str):
        super().__init__("InvalidParameterValue", message)


class ResourceNotFoundException(RESTError):
    def __init__(self, message: str):
        super().__init__("ResourceNotFoundException", message)


class ApplicationNotFound(ElasticBeanstalkException):
    code = 404

    def __init__(self, application_name: str):
        super().__init__(
            "ApplicationNotFound",
            message=f"Elastic Beanstalk application {application_name} not found.",
        )



"""Exceptions raised by the workspaces service."""

from moto.core.exceptions import JsonRESTError


class ValidationException(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("ValidationException", message)


class InvalidParameterValuesException(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("InvalidParameterValuesException", message)


class ResourceAlreadyExistsException(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("ResourceAlreadyExistsException", message)


class ResourceNotFoundException(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("ResourceNotFoundException", message)



import abc
from typing import Any

from moto.stepfunctions.parser.asl.component.common.payload.payloadvalue.payload_value import (
    PayloadValue,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment


class PayloadValueLit(PayloadValue, abc.ABC):
    val: Any

    def __init__(self, val: Any):
        self.val: Any = val

    def _eval_body(self, env: Environment) -> None:
        env.stack.append(self.val)



"""emrcontainers base URL and path."""

from .responses import EMRContainersResponse

url_bases = [
    r"https?://emr-containers\.(.+)\.amazonaws\.com",
]


url_paths = {
    "{0}/virtualclusters$": EMRContainersResponse.dispatch,
    "{0}/virtualclusters/(?P<virtualClusterId>[^/]+)$": EMRContainersResponse.dispatch,
    "{0}/virtualclusters/(?P<virtualClusterId>[^/]+)/jobruns$": EMRContainersResponse.dispatch,
    "{0}/virtualclusters/(?P<virtualClusterId>[^/]+)/jobruns/(?P<jobRunId>[^/]+)$": EMRContainersResponse.dispatch,
}



import abc
import logging
import threading
from typing import Final, Optional

from moto.stepfunctions.parser.api import Timestamp
from moto.stepfunctions.parser.asl.component.program.program import Program
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.utils import TMP_THREADS

LOG = logging.getLogger(__name__)


class BranchWorker:
    class BranchWorkerComm(abc.ABC):
        @abc.abstractmethod
        def on_terminated(self, env: Environment): ...

    _branch_worker_comm: Final[BranchWorkerComm]
    _program: Final[Program]
    _worker_thread: Optional[threading.Thread]
    env: Final[Environment]

    def __init__(
        self, branch_worker_comm: BranchWorkerComm, program: Program, env: Environment
    ):
        self._branch_worker_comm = branch_worker_comm
        self._program = program
        self.env = env
        self._worker_thread = None

    def _thread_routine(self) -> None:
        LOG.info(f"[BranchWorker] [launched] [id: {self._worker_thread.native_id}]")
        self._program.eval(self.env)
        LOG.info(f"[BranchWorker] [terminated] [id: {self._worker_thread.native_id}]")
        self._branch_worker_comm.on_terminated(env=self.env)

    def start(self):
        if self._worker_thread is not None:
            raise RuntimeError(
                f"Attempted to rerun BranchWorker for program ${self._program}."
            )

        self._worker_thread = threading.Thread(
            target=self._thread_routine, name=f"BranchWorker_${self._program}"
        )
        TMP_THREADS.append(self._worker_thread)
        self._worker_thread.start()

    def stop(
        self, stop_date: Timestamp, cause: Optional[str], error: Optional[str]
    ) -> None:
        env = self.env
        if env:
            try:
                env.set_stop(stop_date, cause, error)
            except Exception:
                # Ignore closing exceptions, this method attempts to release resources earlier.
                pass



from moto.stepfunctions.parser.asl.component.common.payload.payloadvalue.payloadtmpl.payload_tmpl import (
    PayloadTmpl,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment


class ResultSelector(EvalComponent):
    def __init__(self, payload_tmpl: PayloadTmpl):
        self.payload_tmpl: PayloadTmpl = payload_tmpl

    def _eval_body(self, env: Environment) -> None:
        self.payload_tmpl.eval(env=env)



from typing import Final

from moto.stepfunctions.parser.asl.component.component import Component


class End(Component):
    def __init__(self, is_end: bool):
        # Designates this state as a terminal state (ends the execution) if set to true.
        self.is_end: Final[bool] = is_end



from .responses import GlacierResponse

url_bases = [r"https?://glacier\.(.+)\.amazonaws.com"]

url_paths = {
    "{0}/(?P<account_number>.+)/vaults$": GlacierResponse.method_dispatch(
        GlacierResponse.all_vault_response
    ),
    "{0}/(?P<account_number>.+)/vaults/(?P<vault_name>[^/]+)$": GlacierResponse.method_dispatch(
        GlacierResponse.vault_response
    ),
    "{0}/(?P<account_number>.+)/vaults/(?P<vault_name>.+)/archives$": GlacierResponse.method_dispatch(
        GlacierResponse.vault_archive_response
    ),
    "{0}/(?P<account_number>.+)/vaults/(?P<vault_name>.+)/archives/(?P<archive_id>.+)$": GlacierResponse.method_dispatch(
        GlacierResponse.vault_archive_individual_response
    ),
    "{0}/(?P<account_number>.+)/vaults/(?P<vault_name>.+)/jobs$": GlacierResponse.method_dispatch(
        GlacierResponse.vault_jobs_response
    ),
    "{0}/(?P<account_number>.+)/vaults/(?P<vault_name>.+)/jobs/(?P<job_id>[^/.]+)$": GlacierResponse.method_dispatch(
        GlacierResponse.vault_jobs_individual_response
    ),
    "{0}/(?P<account_number>.+)/vaults/(?P<vault_name>.+)/jobs/(?P<job_id>.+)/output$": GlacierResponse.method_dispatch(
        GlacierResponse.vault_jobs_output_response
    ),
}



import json
from typing import Any, Dict, List, Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.core.utils import unix_time
from moto.iam.aws_managed_policies import aws_managed_policies_data
from moto.moto_api._internal import mock_random as random
from moto.utilities.paginator import paginate

from .exceptions import (
    ConflictException,
    ResourceNotFoundException,
    ServiceQuotaExceededException,
)
from .utils import PAGINATION_MODEL

# https://docs.aws.amazon.com/singlesignon/latest/userguide/limits.html
MAX_MANAGED_POLICIES_PER_PERMISSION_SET = 20


class AccountAssignment(BaseModel):
    def __init__(
        self,
        instance_arn: str,
        target_id: str,
        target_type: str,
        permission_set_arn: str,
        principal_type: str,
        principal_id: str,
    ):
        self.request_id = str(random.uuid4())
        self.instance_arn = instance_arn
        self.target_id = target_id
        self.target_type = target_type
        self.permission_set_arn = permission_set_arn
        self.principal_type = principal_type
        self.principal_id = principal_id
        self.created_date = unix_time()

    def to_json(self, include_creation_date: bool = False) -> Dict[str, Any]:
        summary: Dict[str, Any] = {
            "TargetId": self.target_id,
            "TargetType": self.target_type,
            "PermissionSetArn": self.permission_set_arn,
            "PrincipalType": self.principal_type,
            "PrincipalId": self.principal_id,
        }
        if include_creation_date:
            summary["CreatedDate"] = self.created_date
        return summary


class PermissionSet(BaseModel):
    def __init__(
        self,
        name: str,
        description: str,
        instance_arn: str,
        session_duration: str,
        relay_state: str,
        tags: List[Dict[str, str]],
    ):
        self.name = name
        self.description = description
        self.instance_arn = instance_arn
        self.permission_set_arn = PermissionSet.generate_id(instance_arn)
        self.session_duration = session_duration
        self.relay_state = relay_state
        self.tags = tags
        self.created_date = unix_time()
        self.inline_policy = ""
        self.managed_policies: List[ManagedPolicy] = list()
        self.customer_managed_policies: List[CustomerManagedPolicy] = list()
        self.total_managed_policies_attached = (
            0  # this will also include customer managed policies
        )

    def to_json(self, include_creation_date: bool = False) -> Dict[str, Any]:
        summary: Dict[str, Any] = {
            "Name": self.name,
            "Description": self.description,
            "PermissionSetArn": self.permission_set_arn,
            "SessionDuration": self.session_duration,
            "RelayState": self.relay_state,
        }
        if include_creation_date:
            summary["CreatedDate"] = self.created_date
        return summary

    @staticmethod
    def generate_id(instance_arn: str) -> str:
        chars = list(range(10)) + ["a", "b", "c", "d", "e", "f"]
        return (
            instance_arn
            + "/ps-"
            + "".join(str(random.choice(chars)) for _ in range(16))
        )


class ManagedPolicy(BaseModel):
    def __init__(self, arn: str, name: str):
        self.arn = arn
        self.name = name

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ManagedPolicy):
            return False
        return self.arn == other.arn


class CustomerManagedPolicy(BaseModel):
    def __init__(self, name: str, path: str = "/"):
        self.name = name
        self.path = path

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, CustomerManagedPolicy):
            return False
        return f"{self.path}{self.name}" == f"{other.path}{other.name}"


class SSOAdminBackend(BaseBackend):
    """Implementation of SSOAdmin APIs."""

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.account_assignments: List[AccountAssignment] = list()
        self.permission_sets: List[PermissionSet] = list()
        self.aws_managed_policies: Optional[Dict[str, Any]] = None

    def create_account_assignment(
        self,
        instance_arn: str,
        target_id: str,
        target_type: str,
        permission_set_arn: str,
        principal_type: str,
        principal_id: str,
    ) -> Dict[str, Any]:
        assignment = AccountAssignment(
            instance_arn,
            target_id,
            target_type,
            permission_set_arn,
            principal_type,
            principal_id,
        )
        self.account_assignments.append(assignment)
        return assignment.to_json()

    def delete_account_assignment(
        self,
        instance_arn: str,
        target_id: str,
        target_type: str,
        permission_set_arn: str,
        principal_type: str,
        principal_id: str,
    ) -> Dict[str, Any]:
        account = self._find_account(
            instance_arn,
            target_id,
            target_type,
            permission_set_arn,
            principal_type,
            principal_id,
        )
        self.account_assignments.remove(account)
        return account.to_json(include_creation_date=True)

    def _find_account(
        self,
        instance_arn: str,
        target_id: str,
        target_type: str,
        permission_set_arn: str,
        principal_type: str,
        principal_id: str,
    ) -> AccountAssignment:
        for account in self.account_assignments:
            instance_arn_match = account.instance_arn == instance_arn
            target_id_match = account.target_id == target_id
            target_type_match = account.target_type == target_type
            permission_set_match = account.permission_set_arn == permission_set_arn
            principal_type_match = account.principal_type == principal_type
            principal_id_match = account.principal_id == principal_id
            if (
                instance_arn_match
                and target_id_match
                and target_type_match
                and permission_set_match
                and principal_type_match
                and principal_id_match
            ):
                return account
        raise ResourceNotFoundException

    def _find_managed_policy(self, managed_policy_arn: str) -> ManagedPolicy:
        """
        Checks to make sure the managed policy exists.
        This pulls from moto/iam/aws_managed_policies.py
        """
        # Lazy loading of aws managed policies file
        if self.aws_managed_policies is None:
            self.aws_managed_policies = json.loads(aws_managed_policies_data)

        policy_name = managed_policy_arn.split("/")[-1]
        managed_policy = self.aws_managed_policies.get(policy_name, None)
        if managed_policy is not None:
            path = managed_policy.get("path", "/")
            expected_arn = f"arn:aws:iam::aws:policy{path}{policy_name}"
            if managed_policy_arn == expected_arn:
                return ManagedPolicy(managed_policy_arn, policy_name)
        raise ResourceNotFoundException(
            f"Policy does not exist with ARN: {managed_policy_arn}"
        )

    @paginate(PAGINATION_MODEL)  # type: ignore[misc]
    def list_account_assignments(
        self, instance_arn: str, account_id: str, permission_set_arn: str
    ) -> List[Dict[str, Any]]:
        account_assignments = []
        for assignment in self.account_assignments:
            if (
                assignment.instance_arn == instance_arn
                and assignment.target_id == account_id
                and assignment.permission_set_arn == permission_set_arn
            ):
                account_assignments.append(
                    {
                        "AccountId": assignment.target_id,
                        "PermissionSetArn": assignment.permission_set_arn,
                        "PrincipalType": assignment.principal_type,
                        "PrincipalId": assignment.principal_id,
                    }
                )
        return account_assignments

    @paginate(PAGINATION_MODEL)  # type: ignore[misc]
    def list_account_assignments_for_principal(
        self,
        filter_: Dict[str, Any],
        instance_arn: str,
        principal_id: str,
        principal_type: str,
    ) -> List[Dict[str, Any]]:
        return [
            {
                "AccountId": account_assignment.target_id,
                "PermissionSetArn": account_assignment.permission_set_arn,
                "PrincipalId": account_assignment.principal_id,
                "PrincipalType": account_assignment.principal_type,
            }
            for account_assignment in self.account_assignments
            if all(
                [
                    filter_.get("AccountId", account_assignment.target_id)
                    == account_assignment.target_id,
                    principal_id == account_assignment.principal_id,
                    principal_type == account_assignment.principal_type,
                    instance_arn == account_assignment.instance_arn,
                ]
            )
        ]

    def create_permission_set(
        self,
        name: str,
        description: str,
        instance_arn: str,
        session_duration: str,
        relay_state: str,
        tags: List[Dict[str, str]],
    ) -> Dict[str, Any]:
        permission_set = PermissionSet(
            name,
            description,
            instance_arn,
            session_duration,
            relay_state,
            tags,
        )
        self.permission_sets.append(permission_set)
        return permission_set.to_json(True)

    def update_permission_set(
        self,
        instance_arn: str,
        permission_set_arn: str,
        description: str,
        session_duration: str,
        relay_state: str,
    ) -> Dict[str, Any]:
        permission_set = self._find_permission_set(
            instance_arn,
            permission_set_arn,
        )
        self.permission_sets.remove(permission_set)
        permission_set.description = description
        permission_set.session_duration = session_duration
        permission_set.relay_state = relay_state
        self.permission_sets.append(permission_set)
        return permission_set.to_json(True)

    def describe_permission_set(
        self, instance_arn: str, permission_set_arn: str
    ) -> Dict[str, Any]:
        permission_set = self._find_permission_set(
            instance_arn,
            permission_set_arn,
        )
        return permission_set.to_json(True)

    def delete_permission_set(
        self, instance_arn: str, permission_set_arn: str
    ) -> Dict[str, Any]:
        permission_set = self._find_permission_set(
            instance_arn,
            permission_set_arn,
        )
        self.permission_sets.remove(permission_set)
        return permission_set.to_json(include_creation_date=True)

    def _find_permission_set(
        self, instance_arn: str, permission_set_arn: str
    ) -> PermissionSet:
        for permission_set in self.permission_sets:
            instance_arn_match = permission_set.instance_arn == instance_arn
            permission_set_match = (
                permission_set.permission_set_arn == permission_set_arn
            )
            if instance_arn_match and permission_set_match:
                return permission_set
        ps_id = permission_set_arn.split("/")[-1]
        raise ResourceNotFoundException(
            message=f"Could not find PermissionSet with id {ps_id}"
        )

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_permission_sets(self, instance_arn: str) -> List[PermissionSet]:
        permission_sets = []
        for permission_set in self.permission_sets:
            if permission_set.instance_arn == instance_arn:
                permission_sets.append(permission_set)
        return permission_sets

    def put_inline_policy_to_permission_set(
        self, instance_arn: str, permission_set_arn: str, inline_policy: str
    ) -> None:
        permission_set = self._find_permission_set(
            instance_arn,
            permission_set_arn,
        )
        permission_set.inline_policy = inline_policy

    def get_inline_policy_for_permission_set(
        self, instance_arn: str, permission_set_arn: str
    ) -> str:
        permission_set = self._find_permission_set(
            instance_arn,
            permission_set_arn,
        )
        return permission_set.inline_policy

    def delete_inline_policy_from_permission_set(
        self, instance_arn: str, permission_set_arn: str
    ) -> None:
        permission_set = self._find_permission_set(
            instance_arn,
            permission_set_arn,
        )
        permission_set.inline_policy = ""

    def attach_managed_policy_to_permission_set(
        self, instance_arn: str, permission_set_arn: str, managed_policy_arn: str
    ) -> None:
        permissionset = self._find_permission_set(
            instance_arn,
            permission_set_arn,
        )
        managed_policy = self._find_managed_policy(managed_policy_arn)

        permissionset_id = permission_set_arn.split("/")[-1]
        if managed_policy in permissionset.managed_policies:
            raise ConflictException(
                f"Permission set with id {permissionset_id} already has a typed link attachment to a manged policy with {managed_policy_arn}"
            )

        if (
            permissionset.total_managed_policies_attached
            >= MAX_MANAGED_POLICIES_PER_PERMISSION_SET
        ):
            permissionset_id = permission_set_arn.split("/")[-1]
            raise ServiceQuotaExceededException(
                f"You have exceeded AWS SSO limits. Cannot create ManagedPolicy more than {MAX_MANAGED_POLICIES_PER_PERMISSION_SET} for id {permissionset_id}. Please refer to https://docs.aws.amazon.com/singlesignon/latest/userguide/limits.html"
            )

        permissionset.managed_policies.append(managed_policy)
        permissionset.total_managed_policies_attached += 1

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_managed_policies_in_permission_set(
        self,
        instance_arn: str,
        permission_set_arn: str,
    ) -> List[ManagedPolicy]:
        permissionset = self._find_permission_set(
            instance_arn,
            permission_set_arn,
        )
        return permissionset.managed_policies

    def _detach_managed_policy(
        self, instance_arn: str, permission_set_arn: str, managed_policy_arn: str
    ) -> None:
        # ensure permission_set exists
        permissionset = self._find_permission_set(
            instance_arn,
            permission_set_arn,
        )

        for managed_policy in permissionset.managed_policies:
            if managed_policy.arn == managed_policy_arn:
                permissionset.managed_policies.remove(managed_policy)
                permissionset.total_managed_policies_attached -= 1
                return

        raise ResourceNotFoundException(
            f"Could not find ManagedPolicy with arn {managed_policy_arn}"
        )

    def detach_managed_policy_from_permission_set(
        self, instance_arn: str, permission_set_arn: str, managed_policy_arn: str
    ) -> None:
        self._detach_managed_policy(
            instance_arn, permission_set_arn, managed_policy_arn
        )

    def attach_customer_managed_policy_reference_to_permission_set(
        self,
        instance_arn: str,
        permission_set_arn: str,
        customer_managed_policy_reference: Dict[str, str],
    ) -> None:
        permissionset = self._find_permission_set(
            permission_set_arn=permission_set_arn, instance_arn=instance_arn
        )

        name = customer_managed_policy_reference["Name"]
        path = customer_managed_policy_reference.get("Path", "/")  # default path is "/"
        customer_managed_policy = CustomerManagedPolicy(name=name, path=path)

        if customer_managed_policy in permissionset.customer_managed_policies:
            raise ConflictException(
                f"Given customer managed policy with name: {name}  and path {path} already attached"
            )

        if (
            permissionset.total_managed_policies_attached
            >= MAX_MANAGED_POLICIES_PER_PERMISSION_SET
        ):
            raise ServiceQuotaExceededException(
                f"Cannot attach managed policy: number of attached managed policies is already at maximum {MAX_MANAGED_POLICIES_PER_PERMISSION_SET}"
            )

        permissionset.customer_managed_policies.append(customer_managed_policy)
        permissionset.total_managed_policies_attached += 1

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_customer_managed_policy_references_in_permission_set(
        self, instance_arn: str, permission_set_arn: str
    ) -> List[CustomerManagedPolicy]:
        permissionset = self._find_permission_set(
            permission_set_arn=permission_set_arn, instance_arn=instance_arn
        )
        return permissionset.customer_managed_policies

    def _detach_customer_managed_policy_from_permissionset(
        self,
        instance_arn: str,
        permission_set_arn: str,
        customer_managed_policy_reference: Dict[str, str],
    ) -> None:
        permissionset = self._find_permission_set(
            permission_set_arn=permission_set_arn, instance_arn=instance_arn
        )
        path: str = customer_managed_policy_reference.get("Path", "/")
        name: str = customer_managed_policy_reference["Name"]

        for customer_managed_policy in permissionset.customer_managed_policies:
            if (
                customer_managed_policy.name == name
                and customer_managed_policy.path == path
            ):
                permissionset.customer_managed_policies.remove(customer_managed_policy)
                permissionset.total_managed_policies_attached -= 1
                return

        raise ResourceNotFoundException(
            f"Given managed policy with name: {name}  and path {path} does not exist on PermissionSet"
        )

    def detach_customer_managed_policy_reference_from_permission_set(
        self,
        instance_arn: str,
        permission_set_arn: str,
        customer_managed_policy_reference: Dict[str, str],
    ) -> None:
        self._detach_customer_managed_policy_from_permissionset(
            instance_arn=instance_arn,
            permission_set_arn=permission_set_arn,
            customer_managed_policy_reference=customer_managed_policy_reference,
        )


ssoadmin_backends = BackendDict(SSOAdminBackend, "sso")



"""ebs base URL and path."""

from .responses import EBSResponse

url_bases = [r"https?://ebs\.(.+)\.amazonaws\.com"]


url_paths = {
    "{0}/snapshots$": EBSResponse.method_dispatch(EBSResponse.snapshots),
    "{0}/snapshots/completion/(?P<snapshot_id>[^/]+)$": EBSResponse.method_dispatch(
        EBSResponse.complete_snapshot
    ),
    "{0}/snapshots/(?P<snapshot_id>[^/]+)/changedblocks$": EBSResponse.method_dispatch(
        EBSResponse.snapshot_changed_blocks
    ),
    "{0}/snapshots/(?P<snapshot_id>[^/]+)/blocks$": EBSResponse.method_dispatch(
        EBSResponse.snapshot_blocks
    ),
    "{0}/snapshots/(?P<snapshot_id>[^/]+)/blocks/(?P<block_idx>[^/]+)$": EBSResponse.method_dispatch(
        EBSResponse.snapshot_block
    ),
}



import json
from json import JSONDecodeError
from typing import Any, Final, Optional

from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.utils.boto_client import boto_client_for
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str


class LambdaFunctionErrorException(Exception):
    function_error: Final[Optional[str]]
    payload: Final[str]

    def __init__(self, function_error: Optional[str], payload: str):
        self.function_error = function_error
        self.payload = payload


def exec_lambda_function(
    env: Environment, parameters: dict, region: str, account: str
) -> None:
    lambda_client = boto_client_for(region=region, account=account, service="lambda")

    invocation_resp = lambda_client.invoke(**parameters)

    func_error: Optional[str] = invocation_resp.get("FunctionError")
    payload_json = json.load(invocation_resp["Payload"])
    if func_error:
        payload_str = json.dumps(payload_json, separators=(",", ":"))
        raise LambdaFunctionErrorException(func_error, payload_str)

    invocation_resp["Payload"] = payload_json

    env.stack.append(invocation_resp)


def to_payload_type(payload: Any) -> Optional[bytes]:
    if isinstance(payload, bytes):
        return payload

    if payload is None:
        str_value = to_json_str(dict())
    elif isinstance(payload, str):
        try:
            json.loads(payload)
            str_value = payload
        except JSONDecodeError:
            str_value = to_json_str(payload)
    else:
        str_value = to_json_str(payload)
    return str_value.encode("utf-8")



import json
from typing import List, Optional

from moto.core.exceptions import JsonRESTError


class RedshiftClientError(JsonRESTError):
    def __init__(self, code: str, message: str):
        super().__init__(error_type=code, message=message)
        self.description = json.dumps(
            {
                "Error": {"Code": code, "Message": message, "Type": "Sender"},
                "RequestId": "6876f774-7273-11e4-85dc-39e55ca848d1",
            }
        )


class ClusterNotFoundError(RedshiftClientError):
    def __init__(self, cluster_identifier: str):
        super().__init__("ClusterNotFound", f"Cluster {cluster_identifier} not found.")


class ClusterSubnetGroupNotFoundError(RedshiftClientError):
    def __init__(self, subnet_identifier: str):
        super().__init__(
            "ClusterSubnetGroupNotFound", f"Subnet group {subnet_identifier} not found."
        )


class ClusterSecurityGroupNotFoundError(RedshiftClientError):
    def __init__(self, group_identifier: str):
        super().__init__(
            "ClusterSecurityGroupNotFound",
            f"Security group {group_identifier} not found.",
        )


class ClusterParameterGroupNotFoundError(RedshiftClientError):
    def __init__(self, group_identifier: str):
        super().__init__(
            "ClusterParameterGroupNotFound",
            f"Parameter group {group_identifier} not found.",
        )


class InvalidSubnetError(RedshiftClientError):
    def __init__(self, subnet_identifier: List[str]):
        super().__init__("InvalidSubnet", f"Subnet {subnet_identifier} not found.")


class SnapshotCopyGrantAlreadyExistsFaultError(RedshiftClientError):
    def __init__(self, snapshot_copy_grant_name: str):
        super().__init__(
            "SnapshotCopyGrantAlreadyExistsFault",
            "Cannot create the snapshot copy grant because a grant "
            f"with the identifier '{snapshot_copy_grant_name}' already exists",
        )


class SnapshotCopyGrantNotFoundFaultError(RedshiftClientError):
    def __init__(self, snapshot_copy_grant_name: str):
        super().__init__(
            "SnapshotCopyGrantNotFoundFault",
            f"Snapshot copy grant not found: {snapshot_copy_grant_name}",
        )


class ClusterSnapshotNotFoundError(RedshiftClientError):
    def __init__(self, snapshot_identifier: str):
        super().__init__(
            "ClusterSnapshotNotFound", f"Snapshot {snapshot_identifier} not found."
        )


class ClusterSnapshotAlreadyExistsError(RedshiftClientError):
    def __init__(self, snapshot_identifier: str):
        super().__init__(
            "ClusterSnapshotAlreadyExists",
            "Cannot create the snapshot because a snapshot with the "
            f"identifier {snapshot_identifier} already exists",
        )


class InvalidParameterValueError(RedshiftClientError):
    def __init__(self, message: str):
        super().__init__("InvalidParameterValue", message)


class ResourceNotFoundFaultError(RedshiftClientError):
    code = 404

    def __init__(
        self,
        resource_type: Optional[str] = None,
        resource_name: Optional[str] = None,
        message: Optional[str] = None,
    ):
        if resource_type and not resource_name:
            msg = f"resource of type '{resource_type}' not found."
        else:
            msg = f"{resource_type} ({resource_name}) not found."
        if message:
            msg = message
        super().__init__("ResourceNotFoundFault", msg)


class SnapshotCopyDisabledFaultError(RedshiftClientError):
    def __init__(self, cluster_identifier: str):
        super().__init__(
            "SnapshotCopyDisabledFault",
            f"Cannot modify retention period because snapshot copy is disabled on Cluster {cluster_identifier}.",
        )


class SnapshotCopyAlreadyDisabledFaultError(RedshiftClientError):
    def __init__(self, cluster_identifier: str):
        super().__init__(
            "SnapshotCopyAlreadyDisabledFault",
            f"Snapshot Copy is already disabled on Cluster {cluster_identifier}.",
        )


class SnapshotCopyAlreadyEnabledFaultError(RedshiftClientError):
    def __init__(self, cluster_identifier: str):
        super().__init__(
            "SnapshotCopyAlreadyEnabledFault",
            f"Snapshot Copy is already enabled on Cluster {cluster_identifier}.",
        )


class ClusterAlreadyExistsFaultError(RedshiftClientError):
    def __init__(self) -> None:
        super().__init__("ClusterAlreadyExists", "Cluster already exists")


class InvalidParameterCombinationError(RedshiftClientError):
    def __init__(self, message: str):
        super().__init__("InvalidParameterCombination", message)


class UnknownSnapshotCopyRegionFaultError(RedshiftClientError):
    def __init__(self, message: str):
        super().__init__("UnknownSnapshotCopyRegionFault", message)


class ClusterSecurityGroupNotFoundFaultError(RedshiftClientError):
    def __init__(self) -> None:
        super().__init__(
            "ClusterSecurityGroupNotFoundFault",
            "The cluster security group name does not refer to an existing cluster security group.",
        )


class InvalidClusterSnapshotStateFaultError(RedshiftClientError):
    def __init__(self, snapshot_identifier: str):
        super().__init__(
            "InvalidClusterSnapshotStateFault",
            f"Cannot delete the snapshot {snapshot_identifier} because only manual snapshots may be deleted",
        )



import re
from io import BytesIO
from typing import Any, Optional, Union

from botocore.awsrequest import AWSResponse

import moto.backend_index as backend_index
from moto.core.base_backend import BackendDict
from moto.core.common_types import TYPE_RESPONSE
from moto.core.config import passthrough_service, passthrough_url
from moto.core.utils import get_equivalent_url_in_aws_domain


class MockRawResponse(BytesIO):
    def __init__(self, response_input: Union[str, bytes]):
        if isinstance(response_input, str):
            response_input = response_input.encode("utf-8")
        super().__init__(response_input)

    def stream(self, **kwargs: Any) -> Any:  # pylint: disable=unused-argument
        contents = self.read()
        while contents:
            yield contents
            contents = self.read()


class BotocoreStubber:
    def __init__(self) -> None:
        self.enabled = False

    def __call__(
        self, event_name: str, request: Any, **kwargs: Any
    ) -> Optional[AWSResponse]:
        if not self.enabled:
            return None

        response = self.process_request(request)
        if response is not None:
            status, headers, body = response
            return AWSResponse(request.url, status, headers, MockRawResponse(body))  # type: ignore[arg-type]
        else:
            return response

    def process_request(self, request: Any) -> Optional[TYPE_RESPONSE]:
        # Handle non-standard AWS endpoint hostnames from ISO regions or custom
        # S3 endpoints.
        parsed_url, _ = get_equivalent_url_in_aws_domain(request.url)
        # Remove the querystring from the URL, as we'll never match on that
        clean_url = f"{parsed_url.scheme}://{parsed_url.netloc}{parsed_url.path}"

        if passthrough_url(clean_url):
            return None

        for service, pattern in backend_index.backend_url_patterns:
            if pattern.match(clean_url):
                if passthrough_service(service):
                    return None

                import moto.backends as backends
                from moto.core import DEFAULT_ACCOUNT_ID
                from moto.core.exceptions import HTTPException

                # TODO: cache this part - we only need backend.urls
                backend_dict = backends.get_backend(service)  # type: ignore[call-overload]

                if isinstance(backend_dict, BackendDict):
                    if "us-east-1" in backend_dict[DEFAULT_ACCOUNT_ID]:
                        backend = backend_dict[DEFAULT_ACCOUNT_ID]["us-east-1"]
                    else:
                        backend = backend_dict[DEFAULT_ACCOUNT_ID]["global"]
                else:
                    backend = backend_dict["global"]

                for header, value in request.headers.items():
                    if isinstance(value, bytes):
                        request.headers[header] = value.decode("utf-8")

                for url, method_to_execute in backend.urls.items():
                    if re.compile(url).match(clean_url):
                        from moto.moto_api import recorder

                        try:
                            recorder._record_request(request)
                            status, headers, body = method_to_execute(
                                request, request.url, request.headers
                            )
                        except HTTPException as e:
                            status = e.code
                            headers = e.get_headers()
                            body = e.get_body()

                        return status, headers, body

        if re.compile(r"https?://.+\.amazonaws.com/.*").match(clean_url):
            return 404, {}, "Not yet implemented"

        return None



from __future__ import annotations

import copy
import logging
import threading
from typing import Any, Dict, Final, List, Optional

from moto.stepfunctions.parser.api import ExecutionFailedEventDetails, Timestamp
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.itemprocessor.map_run_record import (
    MapRunRecordPoolManager,
)
from moto.stepfunctions.parser.asl.eval.aws_execution_details import AWSExecutionDetails
from moto.stepfunctions.parser.asl.eval.callback.callback import CallbackPoolManager
from moto.stepfunctions.parser.asl.eval.contextobject.contex_object import (
    ContextObject,
    ContextObjectInitData,
    ContextObjectManager,
)
from moto.stepfunctions.parser.asl.eval.event.event_history import (
    EventHistory,
    EventHistoryContext,
)
from moto.stepfunctions.parser.asl.eval.program_state import (
    ProgramEnded,
    ProgramError,
    ProgramRunning,
    ProgramState,
    ProgramStopped,
    ProgramTimedOut,
)

LOG = logging.getLogger(__name__)


class Environment:
    _state_mutex: Final[threading.RLock()]
    _program_state: Optional[ProgramState]
    program_state_event: Final[threading.Event()]

    event_history: EventHistory
    event_history_context: Final[EventHistoryContext]
    aws_execution_details: Final[AWSExecutionDetails]
    callback_pool_manager: CallbackPoolManager
    map_run_record_pool_manager: MapRunRecordPoolManager
    context_object_manager: Final[ContextObjectManager]

    _frames: Final[List[Environment]]
    _is_frame: bool = False

    heap: Dict[str, Any] = dict()
    stack: List[Any] = list()
    inp: Optional[Any] = None

    def __init__(
        self,
        aws_execution_details: AWSExecutionDetails,
        context_object_init: ContextObjectInitData,
        event_history_context: EventHistoryContext,
    ):
        super(Environment, self).__init__()
        self._state_mutex = threading.RLock()
        self._program_state = None
        self.program_state_event = threading.Event()

        self.event_history = EventHistory()
        self.event_history_context = event_history_context
        self.aws_execution_details = aws_execution_details
        self.callback_pool_manager = CallbackPoolManager()
        self.map_run_record_pool_manager = MapRunRecordPoolManager()
        self.context_object_manager = ContextObjectManager(
            context_object=ContextObject(
                Execution=context_object_init["Execution"],
                StateMachine=context_object_init["StateMachine"],
            )
        )

        self._frames = list()
        self._is_frame = False

        self.heap = dict()
        self.stack = list()
        self.inp = None

    @classmethod
    def as_frame_of(
        cls, env: Environment, event_history_frame_cache: EventHistoryContext
    ):
        context_object_init = ContextObjectInitData(
            Execution=env.context_object_manager.context_object["Execution"],
            StateMachine=env.context_object_manager.context_object["StateMachine"],
        )
        frame = cls(
            aws_execution_details=env.aws_execution_details,
            context_object_init=context_object_init,
            event_history_context=event_history_frame_cache,
        )
        frame._is_frame = True
        frame.event_history = env.event_history
        if "State" in env.context_object_manager.context_object:
            frame.context_object_manager.context_object["State"] = copy.deepcopy(
                env.context_object_manager.context_object["State"]
            )
        frame.callback_pool_manager = env.callback_pool_manager
        frame.map_run_record_pool_manager = env.map_run_record_pool_manager
        frame.heap = env.heap
        frame._program_state = copy.deepcopy(env._program_state)
        return frame

    @property
    def next_state_name(self) -> Optional[str]:
        next_state_name: Optional[str] = None
        if isinstance(self._program_state, ProgramRunning):
            next_state_name = self._program_state.next_state_name
        return next_state_name

    @next_state_name.setter
    def next_state_name(self, next_state_name: str) -> None:
        if self._program_state is None:
            self._program_state = ProgramRunning()

        if isinstance(self._program_state, ProgramRunning):
            self._program_state.next_state_name = next_state_name
        else:
            raise RuntimeError(
                f"Could not set NextState value when in state '{type(self._program_state)}'."
            )

    def program_state(self) -> ProgramState:
        return copy.deepcopy(self._program_state)

    def is_running(self) -> bool:
        return isinstance(self._program_state, ProgramRunning)

    def set_ended(self) -> None:
        with self._state_mutex:
            if isinstance(self._program_state, ProgramRunning):
                self._program_state = ProgramEnded()
                for frame in self._frames:
                    frame.set_ended()
            self.program_state_event.set()
            self.program_state_event.clear()

    def set_error(self, error: ExecutionFailedEventDetails) -> None:
        with self._state_mutex:
            self._program_state = ProgramError(error=error)
            for frame in self._frames:
                frame.set_error(error=error)
            self.program_state_event.set()
            self.program_state_event.clear()

    def set_timed_out(self) -> None:
        with self._state_mutex:
            self._program_state = ProgramTimedOut()
            for frame in self._frames:
                frame.set_timed_out()
            self.program_state_event.set()
            self.program_state_event.clear()

    def set_stop(
        self, stop_date: Timestamp, cause: Optional[str], error: Optional[str]
    ) -> None:
        with self._state_mutex:
            if isinstance(self._program_state, ProgramRunning):
                self._program_state = ProgramStopped(
                    stop_date=stop_date, cause=cause, error=error
                )
                for frame in self._frames:
                    frame.set_stop(stop_date=stop_date, cause=cause, error=error)
                self.program_state_event.set()
                self.program_state_event.clear()
            else:
                raise RuntimeError("Cannot stop non running ProgramState.")

    def open_frame(
        self, event_history_context: Optional[EventHistoryContext] = None
    ) -> Environment:
        with self._state_mutex:
            # The default logic provisions for child frame to extend the source frame event id.
            if event_history_context is None:
                event_history_context = EventHistoryContext(
                    previous_event_id=self.event_history_context.source_event_id
                )

            frame = Environment.as_frame_of(self, event_history_context)
            self._frames.append(frame)
            return frame

    def close_frame(self, frame: Environment) -> None:
        with self._state_mutex:
            if frame in self._frames:
                self._frames.remove(frame)
                self.event_history_context.integrate(frame.event_history_context)

    def delete_frame(self, frame: Environment) -> None:
        with self._state_mutex:
            if frame in self._frames:
                self._frames.remove(frame)

    def is_frame(self) -> bool:
        return self._is_frame



import hashlib
from collections import OrderedDict
from typing import Any, Dict, List

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel

from .exceptions import ClientError


class Object(BaseModel):
    def __init__(
        self, path: str, body: str, etag: str, storage_class: str = "TEMPORAL"
    ):
        self.path = path
        self.body = body
        self.content_sha256 = hashlib.sha256(body.encode("utf-8")).hexdigest()
        self.etag = etag
        self.storage_class = storage_class

    def to_dict(self) -> Dict[str, Any]:
        return {
            "ETag": self.etag,
            "Name": self.path,
            "Type": "FILE",
            "ContentLength": 123,
            "StorageClass": self.storage_class,
            "Path": self.path,
            "ContentSHA256": self.content_sha256,
        }


class MediaStoreDataBackend(BaseBackend):
    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self._objects: Dict[str, Object] = OrderedDict()

    def put_object(
        self, body: str, path: str, storage_class: str = "TEMPORAL"
    ) -> Object:
        """
        The following parameters are not yet implemented: ContentType, CacheControl, UploadAvailability
        """
        new_object = Object(
            path=path, body=body, etag="etag", storage_class=storage_class
        )
        self._objects[path] = new_object
        return new_object

    def delete_object(self, path: str) -> None:
        if path not in self._objects:
            raise ClientError(
                "ObjectNotFoundException", f"Object with id={path} not found"
            )
        del self._objects[path]

    def get_object(self, path: str) -> Object:
        """
        The Range-parameter is not yet supported.
        """
        objects_found = [item for item in self._objects.values() if item.path == path]
        if len(objects_found) == 0:
            raise ClientError(
                "ObjectNotFoundException", f"Object with id={path} not found"
            )
        return objects_found[0]

    def list_items(self) -> List[Dict[str, Any]]:
        """
        The Path- and MaxResults-parameters are not yet supported.
        """
        return [c.to_dict() for c in self._objects.values()]


mediastoredata_backends = BackendDict(MediaStoreDataBackend, "mediastore-data")



"""Handles incoming signer requests, invokes methods, returns responses."""

import json
from typing import Any
from urllib.parse import unquote

from moto.core.responses import BaseResponse

from .models import SignerBackend, signer_backends


class SignerResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="signer")

    @property
    def signer_backend(self) -> SignerBackend:
        """Return backend instance specific for this region."""
        return signer_backends[self.current_account][self.region]

    def cancel_signing_profile(self) -> str:
        profile_name = self.path.split("/")[-1]
        self.signer_backend.cancel_signing_profile(profile_name=profile_name)
        return "{}"

    def get_signing_profile(self) -> str:
        profile_name = self.path.split("/")[-1]
        profile = self.signer_backend.get_signing_profile(profile_name=profile_name)
        return json.dumps(profile.to_dict())

    def put_signing_profile(self) -> str:
        params = json.loads(self.body)
        profile_name = self.path.split("/")[-1]
        signature_validity_period = params.get("signatureValidityPeriod")
        platform_id = params.get("platformId")
        tags = params.get("tags")
        signing_material = params.get("signingMaterial")
        profile = self.signer_backend.put_signing_profile(
            profile_name=profile_name,
            signature_validity_period=signature_validity_period,
            platform_id=platform_id,
            signing_material=signing_material,
            tags=tags,
        )
        return json.dumps(profile.to_dict(full=False))

    def list_signing_platforms(self) -> str:
        platforms = self.signer_backend.list_signing_platforms()
        return json.dumps(dict(platforms=platforms))

    def list_tags_for_resource(self) -> str:
        resource_arn = unquote(self.path.split("/tags/")[-1])
        return json.dumps(
            {"tags": self.signer_backend.list_tags_for_resource(resource_arn)}
        )

    def tag_resource(self) -> str:
        resource_arn = unquote(self.path.split("/tags/")[-1])
        tags = self._get_param("tags")
        self.signer_backend.tag_resource(resource_arn, tags)
        return "{}"

    def untag_resource(self) -> str:
        resource_arn = unquote(self.path.split("/tags/")[-1])
        tag_keys = self.querystring.get("tagKeys")
        self.signer_backend.untag_resource(resource_arn, tag_keys)  # type: ignore
        return "{}"

    def tags(self, request: Any, full_url: str, headers: Any) -> str:  # type: ignore[return]
        self.setup_class(request, full_url, headers)
        if request.method == "GET":
            return self.list_tags_for_resource()
        if request.method == "POST":
            return self.tag_resource()
        if request.method == "DELETE":
            return self.untag_resource()



import re
from datetime import datetime, timedelta, timezone
from enum import Enum
from ipaddress import IPv4Address, IPv6Address, ip_address
from typing import Any, Dict, List, Optional, Type

from moto.core.common_models import BaseModel
from moto.moto_api._internal import mock_random
from moto.route53domains.exceptions import UnsupportedTLDException

DOMAIN_OPERATION_STATUSES = (
    "SUBMITTED",
    "IN_PROGRESS",
    "ERROR",
    "SUCCESSFUL",
    "FAILED",
)

DOMAIN_OPERATION_TYPES = (
    "REGISTER_DOMAIN",
    "DELETE_DOMAIN",
    "TRANSFER_IN_DOMAIN",
    "UPDATE_DOMAIN_CONTACT",
    "UPDATE_NAMESERVER",
    "CHANGE_PRIVACY_PROTECTION",
    "DOMAIN_LOCK",
    "ENABLE_AUTORENEW",
    "DISABLE_AUTORENEW",
    "ADD_DNSSEC",
    "REMOVE_DNSSEC",
    "EXPIRE_DOMAIN",
    "TRANSFER_OUT_DOMAIN",
    "CHANGE_DOMAIN_OWNER",
    "RENEW_DOMAIN",
    "PUSH_DOMAIN",
    "INTERNAL_TRANSFER_OUT_DOMAIN",
    "INTERNAL_TRANSFER_IN_DOMAIN",
)

DOMAIN_OPERATION_STATUS_FLAGS = (
    "PENDING_ACCEPTANCE",
    "PENDING_CUSTOMER_ACTION",
    "PENDING_AUTHORIZATION",
    "PENDING_PAYMENT_VERIFICATION",
    "PENDING_SUPPORT_CASE",
)

DOMAIN_CONTACT_DETAIL_CONTACT_TYPES = (
    "PERSON",
    "COMPANY",
    "ASSOCIATION",
    "PUBLIC_BODY",
    "RESELLER",
    "ORGANIZATION",
)

DOMAIN_CONTACT_DETAIL_COUNTRY_CODES = (
    "AC",
    "AD",
    "AE",
    "AF",
    "AG",
    "AI",
    "AL",
    "AM",
    "AN",
    "AO",
    "AQ",
    "AR",
    "AS",
    "AT",
    "AU",
    "AW",
    "AX",
    "AZ",
    "BA",
    "BB",
    "BD",
    "BE",
    "BF",
    "BG",
    "BH",
    "BI",
    "BJ",
    "BL",
    "BM",
    "BN",
    "BO",
    "BQ",
    "BR",
    "BS",
    "BT",
    "BV",
    "BW",
    "BY",
    "BZ",
    "CA",
    "CC",
    "CD",
    "CF",
    "CG",
    "CH",
    "CI",
    "CK",
    "CL",
    "CM",
    "CN",
    "CO",
    "CR",
    "CU",
    "CV",
    "CW",
    "CX",
    "CY",
    "CZ",
    "DE",
    "DJ",
    "DK",
    "DM",
    "DO",
    "DZ",
    "EC",
    "EE",
    "EG",
    "EH",
    "ER",
    "ES",
    "ET",
    "FI",
    "FJ",
    "FK",
    "FM",
    "FO",
    "FR",
    "GA",
    "GB",
    "GD",
    "GE",
    "GF",
    "GG",
    "GH",
    "GI",
    "GL",
    "GM",
    "GN",
    "GP",
    "GQ",
    "GR",
    "GS",
    "GT",
    "GU",
    "GW",
    "GY",
    "HK",
    "HM",
    "HN",
    "HR",
    "HT",
    "HU",
    "ID",
    "IE",
    "IL",
    "IM",
    "IN",
    "IO",
    "IQ",
    "IR",
    "IS",
    "IT",
    "JE",
    "JM",
    "JO",
    "JP",
    "KE",
    "KG",
    "KH",
    "KI",
    "KM",
    "KN",
    "KP",
    "KR",
    "KW",
    "KY",
    "KZ",
    "LA",
    "LB",
    "LC",
    "LI",
    "LK",
    "LR",
    "LS",
    "LT",
    "LU",
    "LV",
    "LY",
    "MA",
    "MC",
    "MD",
    "ME",
    "MF",
    "MG",
    "MH",
    "MK",
    "ML",
    "MM",
    "MN",
    "MO",
    "MP",
    "MQ",
    "MR",
    "MS",
    "MT",
    "MU",
    "MV",
    "MW",
    "MX",
    "MY",
    "MZ",
    "NA",
    "NC",
    "NE",
    "NF",
    "NG",
    "NI",
    "NL",
    "NO",
    "NP",
    "NR",
    "NU",
    "NZ",
    "OM",
    "PA",
    "PE",
    "PF",
    "PG",
    "PH",
    "PK",
    "PL",
    "PM",
    "PN",
    "PR",
    "PS",
    "PT",
    "PW",
    "PY",
    "QA",
    "RE",
    "RO",
    "RS",
    "RU",
    "RW",
    "SA",
    "SB",
    "SC",
    "SD",
    "SE",
    "SG",
    "SH",
    "SI",
    "SJ",
    "SK",
    "SL",
    "SM",
    "SN",
    "SO",
    "SR",
    "SS",
    "ST",
    "SV",
    "SX",
    "SY",
    "SZ",
    "TC",
    "TD",
    "TF",
    "TG",
    "TH",
    "TJ",
    "TK",
    "TL",
    "TM",
    "TN",
    "TO",
    "TP",
    "TR",
    "TT",
    "TV",
    "TW",
    "TZ",
    "UA",
    "UG",
    "US",
    "UY",
    "UZ",
    "VA",
    "VC",
    "VE",
    "VG",
    "VI",
    "VN",
    "VU",
    "WF",
    "WS",
    "YE",
    "YT",
    "ZA",
    "ZM",
    "ZW",
)

# List of supported top-level domains that you can register with Amazon Route53
AWS_SUPPORTED_TLDS = (
    "ac",
    "academy",
    "accountants",
    "actor",
    "adult",
    "agency",
    "airforce",
    "apartments",
    "associates",
    "auction",
    "audio",
    "band",
    "bargains",
    "bet",
    "bike",
    "bingo",
    "biz",
    "black",
    "blue",
    "boutique",
    "builders",
    "business",
    "buzz",
    "cab",
    "cafe",
    "camera",
    "camp",
    "capital",
    "cards",
    "care",
    "careers",
    "cash",
    "casino",
    "catering",
    "cc",
    "center",
    "ceo",
    "chat",
    "cheap",
    "church",
    "city",
    "claims",
    "cleaning",
    "click",
    "clinic",
    "clothing",
    "cloud",
    "club",
    "coach",
    "codes",
    "coffee",
    "college",
    "com",
    "community",
    "company",
    "computer",
    "condos",
    "construction",
    "consulting",
    "contractors",
    "cool",
    "coupons",
    "credit",
    "creditcard",
    "cruises",
    "dance",
    "dating",
    "deals",
    "degree",
    "delivery",
    "democrat",
    "dental",
    "diamonds",
    "diet",
    "digital",
    "direct",
    "directory",
    "discount",
    "dog",
    "domains",
    "education",
    "email",
    "energy",
    "engineering",
    "enterprises",
    "equipment",
    "estate",
    "events",
    "exchange",
    "expert",
    "exposed",
    "express",
    "fail",
    "farm",
    "finance",
    "financial",
    "fish",
    "fitness",
    "flights",
    "florist",
    "flowers",
    "fm",
    "football",
    "forsale",
    "foundation",
    "fund",
    "furniture",
    "futbol",
    "fyi",
    "gallery",
    "games",
    "gift",
    "gifts",
    "gives",
    "glass",
    "global",
    "gmbh",
    "gold",
    "golf",
    "graphics",
    "gratis",
    "green",
    "gripe",
    "group",
    "guide",
    "guitars",
    "guru",
    "haus",
    "healthcare",
    "help",
    "hiv",
    "hockey",
    "holdings",
    "holiday",
    "host",
    "hosting",
    "house",
    "im",
    "immo",
    "immobilien",
    "industries",
    "info",
    "ink",
    "institute",
    "insure",
    "international",
    "investments",
    "io",
    "irish",
    "jewelry",
    "juegos",
    "kaufen",
    "kim",
    "kitchen",
    "kiwi",
    "land",
    "lease",
    "legal",
    "lgbt",
    "life",
    "lighting",
    "limited",
    "limo",
    "link",
    "live",
    "loan",
    "loans",
    "lol",
    "maison",
    "management",
    "marketing",
    "mba",
    "media",
    "memorial",
    "mobi",
    "moda",
    "money",
    "mortgage",
    "movie",
    "name",
    "net",
    "network",
    "news",
    "ninja",
    "onl",
    "online",
    "org",
    "partners",
    "parts",
    "photo",
    "photography",
    "photos",
    "pics",
    "pictures",
    "pink",
    "pizza",
    "place",
    "plumbing",
    "plus",
    "poker",
    "porn",
    "press",
    "pro",
    "productions",
    "properties",
    "property",
    "pub",
    "qpon",
    "recipes",
    "red",
    "reise",
    "reisen",
    "rentals",
    "repair",
    "report",
    "republican",
    "restaurant",
    "reviews",
    "rip",
    "rocks",
    "run",
    "sale",
    "sarl",
    "school",
    "schule",
    "services",
    "sex",
    "sexy",
    "shiksha",
    "shoes",
    "show",
    "singles",
    "site",
    "soccer",
    "social",
    "solar",
    "solutions",
    "space",
    "store",
    "studio",
    "style",
    "sucks",
    "supplies",
    "supply",
    "support",
    "surgery",
    "systems",
    "tattoo",
    "tax",
    "taxi",
    "team",
    "tech",
    "technology",
    "tennis",
    "theater",
    "tienda",
    "tips",
    "tires",
    "today",
    "tools",
    "tours",
    "town",
    "toys",
    "trade",
    "training",
    "tv",
    "university",
    "uno",
    "vacations",
    "vegas",
    "ventures",
    "vg",
    "viajes",
    "video",
    "villas",
    "vision",
    "voyage",
    "watch",
    "website",
    "wedding",
    "wiki",
    "wine",
    "works",
    "world",
    "wtf",
    "xyz",
    "zone",
)

VALID_DOMAIN_REGEX = re.compile(
    r"(?:[a-z0-9](?:[a-z0-9-]{0,61}[a-z0-9])?\.)+[a-z0-9][a-z0-9-]{0,61}[a-z0-9]"
)
PHONE_NUMBER_REGEX = re.compile(r"\+\d*\.\d+$")


class DomainFilterField(str, Enum):
    DOMAIN_NAME = "DomainName"
    EXPIRY = "Expiry"


class DomainSortOrder(str, Enum):
    ASCENDING = "ASC"
    DESCENDING = "DES"


class DomainFilterOperator(str, Enum):
    LE = "LE"
    GE = "GE"
    BEGINS_WITH = "BEGINS_WITH"


def is_valid_enum(value: Any, enum_cls: Type[Enum]) -> bool:
    try:
        enum_cls(value)
        return True
    except ValueError:
        return False


class ValidationException(Exception):
    def __init__(self, errors: List[str]):
        super().__init__("\n\t".join(errors))
        self.errors = errors


class Route53DomainsOperation(BaseModel):
    def __init__(
        self,
        id_: str,
        domain_name: str,
        status: str,
        type_: str,
        submitted_date: datetime,
        last_updated_date: datetime,
        message: Optional[str] = None,
        status_flag: Optional[str] = None,
    ):
        self.id = id_
        self.domain_name = domain_name
        self.status = status
        self.type = type_
        self.submitted_date = submitted_date
        self.last_updated_date = last_updated_date
        self.message = message
        self.status_flag = status_flag

    @classmethod
    def validate(  # type: ignore[misc,no-untyped-def]
        cls,
        domain_name: str,
        status: str,
        type_: str,
        message: Optional[str] = None,
        status_flag: Optional[str] = None,
    ):
        id_ = str(mock_random.uuid4())
        submitted_date = datetime.now(timezone.utc)
        last_updated_date = datetime.now(timezone.utc)

        return cls(
            id_,
            domain_name,
            status,
            type_,
            submitted_date,
            last_updated_date,
            message,
            status_flag,
        )

    def to_json(self) -> Dict[str, Any]:
        d = {
            "OperationId": self.id,
            "Status": self.status,
            "StatusFlag": self.status_flag,
            "DomainName": self.domain_name,
            "LastUpdatedDate": self.last_updated_date.timestamp(),
            "SubmittedDate": self.submitted_date.timestamp(),
            "Type": self.type,
        }

        if self.message:
            d["Message"] = self.message

        return d


class Route53DomainsContactDetail(BaseModel):
    def __init__(
        self,
        address_line_1: Optional[str] = None,
        address_line_2: Optional[str] = None,
        city: Optional[str] = None,
        contact_type: Optional[str] = None,
        country_code: Optional[str] = None,
        email: Optional[str] = None,
        extra_params: Optional[List[Dict[str, Any]]] = None,
        fax: Optional[str] = None,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None,
        organization_name: Optional[str] = None,
        phone_number: Optional[str] = None,
        state: Optional[str] = None,
        zip_code: Optional[str] = None,
    ):
        super().__init__()
        self.address_line_1 = address_line_1
        self.address_line_2 = address_line_2
        self.city = city
        self.contact_type = contact_type
        self.country_code = country_code
        self.email = email
        self.extra_params = extra_params
        self.fax = fax
        self.first_name = first_name
        self.last_name = last_name
        self.organization_name = organization_name
        self.phone_number = phone_number
        self.state = state
        self.zip_code = zip_code

    @classmethod
    def validate(  # type: ignore[misc, no-untyped-def]
        cls,
        address_line_1: Optional[str] = None,
        address_line_2: Optional[str] = None,
        city: Optional[str] = None,
        contact_type: Optional[str] = None,
        country_code: Optional[str] = None,
        email: Optional[str] = None,
        extra_params: Optional[List[Dict[str, Any]]] = None,
        fax: Optional[str] = None,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None,
        organization_name: Optional[str] = None,
        phone_number: Optional[str] = None,
        state: Optional[str] = None,
        zip_code: Optional[str] = None,
    ):
        input_errors: List[str] = []

        cls.__validate_str_len(address_line_1, "AddressLine1", 255, input_errors)
        cls.__validate_str_len(address_line_2, "AddressLine2", 255, input_errors)
        cls.__validate_str_len(city, "City", 255, input_errors)
        cls.__validate_str_len(email, "Email", 255, input_errors)
        cls.__validate_str_len(fax, "Fax", 255, input_errors)
        cls.__validate_str_len(first_name, "FirstName", 255, input_errors)
        cls.__validate_str_len(last_name, "LastName", 255, input_errors)
        cls.__validate_str_len(state, "State", 255, input_errors)
        cls.__validate_str_len(zip_code, "ZipCode", 255, input_errors)

        if contact_type:
            if contact_type not in DOMAIN_CONTACT_DETAIL_CONTACT_TYPES:
                input_errors.append(f"Invalid contact type {contact_type}")
            else:
                if contact_type != "PERSON" and not organization_name:
                    input_errors.append(
                        "Must supply OrganizationName when ContactType is not PERSON"
                    )

        if country_code and country_code not in DOMAIN_CONTACT_DETAIL_COUNTRY_CODES:
            input_errors.append(f"CountryCode {country_code} is invalid")

        if phone_number and not PHONE_NUMBER_REGEX.match(phone_number):
            input_errors.append("PhoneNumber is in an invalid format")

        if input_errors:
            raise ValidationException(input_errors)

        return cls(
            address_line_1,
            address_line_2,
            city,
            contact_type,
            country_code,
            email,
            extra_params,
            fax,
            first_name,
            last_name,
            organization_name,
            phone_number,
            state,
            zip_code,
        )

    @classmethod
    def validate_dict(cls, d: Dict[str, Any]):  # type: ignore[misc, no-untyped-def]
        address_line_1 = d.get("AddressLine1")
        address_line_2 = d.get("AddressLine2")
        city = d.get("City")
        contact_type = d.get("ContactType")
        country_code = d.get("CountryCode")
        email = d.get("Email")
        extra_params = d.get("ExtraParams")
        fax = d.get("Fax")
        first_name = d.get("FirstName")
        last_name = d.get("LastName")
        organization_name = d.get("OrganizationName")
        phone_number = d.get("PhoneNumber")
        state = d.get("State")
        zip_code = d.get("ZipCode")
        return cls.validate(
            address_line_1=address_line_1,
            address_line_2=address_line_2,
            city=city,
            contact_type=contact_type,
            country_code=country_code,
            email=email,
            extra_params=extra_params,
            fax=fax,
            first_name=first_name,
            last_name=last_name,
            organization_name=organization_name,
            phone_number=phone_number,
            state=state,
            zip_code=zip_code,
        )

    @staticmethod
    def __validate_str_len(
        value: Optional[str], field_name: str, max_len: int, input_errors: List[str]
    ) -> None:
        if value and len(value) > max_len:
            input_errors.append(f"Length of {field_name} is more than {max_len}")

    def to_json(self) -> Dict[str, Any]:
        d = {
            "FirstName": self.first_name,
            "LastName": self.last_name,
            "ContactType": self.contact_type,
            "OrganizationName": self.organization_name,
            "AddressLine1": self.address_line_1,
            "AddressLine2": self.address_line_2,
            "City": self.city,
            "State": self.state,
            "CountryCode": self.country_code,
            "ZipCode": self.zip_code,
            "PhoneNumber": self.phone_number,
            "Email": self.email,
            "Fax": self.fax,
            "ExtraParams": self.extra_params,
        }

        return {key: value for key, value in d.items() if value is not None}


class NameServer:
    def __init__(self, name: str, glue_ips: List[str]):
        self.name = name
        self.glue_ips = glue_ips

    @classmethod
    def validate(cls, name: str, glue_ips: Optional[List[str]] = None):  # type: ignore[misc,no-untyped-def]
        glue_ips = glue_ips or []
        input_errors: List[str] = []

        if not VALID_DOMAIN_REGEX.match(name):
            input_errors.append(f"{name} is not a valid host name")

        num_ipv4_addresses = 0
        num_ipv6_addresses = 0
        for ip in glue_ips:
            try:
                address = ip_address(ip)
                if isinstance(address, IPv4Address):
                    num_ipv4_addresses += 1
                elif isinstance(address, IPv6Address):
                    num_ipv6_addresses += 1
            except ValueError:
                input_errors.append(f"{ip} is not a valid IP address")

        if num_ipv4_addresses > 1:
            input_errors.append("GlueIps list must include only 1 IPv4 address")
        if num_ipv6_addresses > 1:
            input_errors.append("GlueIps list must include only 1 IPv6 address")

        if input_errors:
            raise ValidationException(input_errors)

        return cls(name, glue_ips)

    @classmethod
    def validate_dict(cls, data: Dict[str, Any]):  # type: ignore[misc,no-untyped-def]
        name = data.get("Name")
        glue_ips = data.get("GlueIps")
        return cls.validate(name, glue_ips)  # type: ignore[arg-type]

    def to_json(self) -> Dict[str, Any]:
        d: Dict[str, Any] = {"Name": self.name}
        if self.glue_ips:
            d["GlueIps"] = self.glue_ips
        return d


class Route53Domain(BaseModel):
    def __init__(
        self,
        domain_name: str,
        nameservers: List[NameServer],
        auto_renew: bool,
        admin_contact: Route53DomainsContactDetail,
        registrant_contact: Route53DomainsContactDetail,
        tech_contact: Route53DomainsContactDetail,
        admin_privacy: bool,
        registrant_privacy: bool,
        tech_privacy: bool,
        registrar_name: str,
        whois_server: str,
        registrar_url: str,
        abuse_contact_email: str,
        abuse_contact_phone: str,
        registry_domain_id: str,
        creation_date: datetime,
        updated_date: datetime,
        expiration_date: datetime,
        reseller: str,
        status_list: List[str],
        dns_sec_keys: List[Dict[str, Any]],
        extra_params: List[Dict[str, Any]],
    ):
        self.domain_name = domain_name
        self.nameservers = nameservers
        self.auto_renew = auto_renew
        self.admin_contact = admin_contact
        self.registrant_contact = registrant_contact
        self.tech_contact = tech_contact
        self.admin_privacy = admin_privacy
        self.registrant_privacy = registrant_privacy
        self.tech_privacy = tech_privacy
        self.registrar_name = registrar_name
        self.whois_server = whois_server
        self.registrar_url = registrar_url
        self.abuse_contact_email = abuse_contact_email
        self.abuse_contact_phone = abuse_contact_phone
        self.registry_domain_id = registry_domain_id
        self.creation_date = creation_date
        self.updated_date = updated_date
        self.expiration_date = expiration_date
        self.reseller = reseller
        self.status_list = status_list
        self.dns_sec_keys = dns_sec_keys
        self.extra_params = extra_params

    @classmethod
    def validate(  # type: ignore[misc,no-untyped-def]
        cls,
        domain_name: str,
        admin_contact: Route53DomainsContactDetail,
        registrant_contact: Route53DomainsContactDetail,
        tech_contact: Route53DomainsContactDetail,
        nameservers: Optional[List[Dict[str, Any]]] = None,  # type: ignore[type-arg]
        auto_renew: bool = True,
        admin_privacy: bool = True,
        registrant_privacy: bool = True,
        tech_privacy: bool = True,
        registrar_name: Optional[str] = None,
        whois_server: Optional[str] = None,
        registrar_url: Optional[str] = None,
        abuse_contact_email: Optional[str] = None,
        abuse_contact_phone: Optional[str] = None,
        registry_domain_id: Optional[str] = None,
        expiration_date: Optional[datetime] = None,
        reseller: Optional[str] = None,
        dns_sec_keys: Optional[List[Dict[str, Any]]] = None,
        extra_params: Optional[List[Dict[str, Any]]] = None,
    ):
        input_errors: List[str] = []

        cls.validate_domain_name(domain_name, input_errors)

        nameservers = nameservers or []
        try:
            nameservers = [
                NameServer.validate_dict(nameserver) for nameserver in nameservers
            ] or [
                NameServer.validate(name="ns-2048.awscdn-64.net"),
                NameServer.validate(name="ns-2051.awscdn-67.net"),
                NameServer.validate(name="ns-2050.awscdn-66.net"),
                NameServer.validate(name="ns-2049.awscdn-65.net"),
            ]
        except ValidationException as e:
            input_errors += e.errors

        creation_date = datetime.now(timezone.utc)
        updated_date = datetime.now(timezone.utc)
        expiration_date = expiration_date or datetime.now(timezone.utc) + timedelta(
            days=365 * 10
        )
        registrar_name = registrar_name or "GANDI SAS"
        whois_server = whois_server or "whois.gandi.net"
        registrar_url = registrar_url or "http://www.gandi.net"
        abuse_contact_email = abuse_contact_email or "abuse@support.gandi.net"
        status_list = ["SUCCEEDED"]

        time_until_expiration = expiration_date - datetime.now(timezone.utc)
        if time_until_expiration < timedelta(
            days=365
        ) or time_until_expiration > timedelta(days=365 * 10):
            input_errors.append(
                "ExpirationDate must by between 1 and 10 years from now"
            )

        if input_errors:
            raise ValidationException(input_errors)

        return cls(
            domain_name=domain_name,
            nameservers=nameservers,  # type: ignore[arg-type]
            auto_renew=auto_renew,  # type: ignore[arg-type]
            admin_contact=admin_contact,
            registrant_contact=registrant_contact,
            tech_contact=tech_contact,
            admin_privacy=admin_privacy,
            registrant_privacy=registrant_privacy,
            tech_privacy=tech_privacy,
            registrar_name=registrar_name,
            whois_server=whois_server,
            registrar_url=registrar_url,
            abuse_contact_email=abuse_contact_email,  # type: ignore[arg-type]
            abuse_contact_phone=abuse_contact_phone,  # type: ignore[arg-type]
            registry_domain_id=registry_domain_id,  # type: ignore[arg-type]
            creation_date=creation_date,  # type: ignore[arg-type]
            updated_date=updated_date,
            expiration_date=expiration_date,
            reseller=reseller,  # type: ignore[arg-type]
            status_list=status_list,  # type: ignore[arg-type]
            dns_sec_keys=dns_sec_keys,  # type: ignore[arg-type]
            extra_params=extra_params,  # type: ignore[arg-type]
        )

    @staticmethod
    def validate_domain_name(domain_name: str, input_errors: List[str]) -> None:
        if not VALID_DOMAIN_REGEX.match(domain_name):
            input_errors.append("Invalid domain name")
            return

        tld = domain_name.split(".")[-1]
        if tld not in AWS_SUPPORTED_TLDS:
            raise UnsupportedTLDException(tld)

    def to_json(self) -> Dict[str, Any]:
        return {
            "DomainName": self.domain_name,
            "Nameservers": [nameserver.to_json() for nameserver in self.nameservers],
            "AutoRenew": self.auto_renew,
            "AdminContact": self.admin_contact.to_json(),
            "RegistrantContact": self.registrant_contact.to_json(),
            "TechContact": self.tech_contact.to_json(),
            "AdminPrivacy": self.admin_privacy,
            "RegistrantPrivacy": self.registrant_privacy,
            "TechPrivacy": self.tech_privacy,
            "RegistrarName": self.registrar_name,
            "WhoIsServer": self.whois_server,
            "RegistrarUrl": self.registrar_url,
            "AbuseContactEmail": self.abuse_contact_email,
            "AbuseContactPhone": self.abuse_contact_phone,
            "RegistryDomainId": "",
            "CreationDate": self.creation_date.timestamp(),
            "UpdateDate": self.updated_date.timestamp(),
            "ExpirationDate": self.expiration_date.timestamp(),
            "Reseller": self.reseller,
            "DnsSec": "",
            "StatusList": self.status_list,
            "DnsSecKeys": self.dns_sec_keys,
            "BillingContact": self.admin_contact.to_json(),
        }


class DomainsFilter:
    def __init__(
        self, name: DomainFilterField, operator: DomainFilterOperator, values: List[str]
    ):
        self.name: DomainFilterField = name
        self.operator: DomainFilterOperator = operator
        self.values = values

    def filter(self, domain: Route53Domain) -> bool:
        if self.name == DomainFilterField.DOMAIN_NAME:
            return self.__filter_by_domain_name(domain)
        return self.__filter_by_expiry_date(domain)

    def __filter_by_domain_name(self, domain: Route53Domain) -> bool:
        return any([domain.domain_name.startswith(value) for value in self.values])

    def __filter_by_expiry_date(self, domain: Route53Domain) -> bool:
        return (
            any(
                [
                    value
                    for value in self.values
                    if domain.expiration_date
                    >= datetime.fromtimestamp(float(value), tz=timezone.utc)
                ]
            )
            if self.operator == DomainFilterOperator.GE
            else any(
                [
                    value
                    for value in self.values
                    if domain.expiration_date
                    <= datetime.fromtimestamp(float(value), tz=timezone.utc)
                ]
            )
        )

    @classmethod
    def validate(cls, name: str, operator: str, values: List[str]):  # type: ignore[misc, no-untyped-def]
        input_errors: List[str] = []

        if not is_valid_enum(name, DomainFilterField):
            input_errors.append(f"Cannot filter by field {name}")

        if not is_valid_enum(operator, DomainFilterOperator):
            input_errors.append(f"Invalid filter operator {operator}")

        if len(values) != 1:
            input_errors.append("Multiple filter values are not currently supported")

        if (
            name == DomainFilterField.DOMAIN_NAME
            and operator != DomainFilterOperator.BEGINS_WITH
        ):
            input_errors.append(
                f"Operator {operator} cannot be used with the DomainName filter"
            )

        if name == DomainFilterField.EXPIRY and operator not in (
            DomainFilterOperator.GE,
            DomainFilterOperator.LE,
        ):
            input_errors.append(
                f"Operator {operator} cannot be used with the Expiry filter"
            )

        if input_errors:
            raise ValidationException(input_errors)

        return cls(
            name=DomainFilterField(name),
            operator=DomainFilterOperator(operator),
            values=values,
        )

    @classmethod
    def validate_dict(cls, data: Dict[str, Any]):  # type: ignore[misc,no-untyped-def]
        name = data.get("Name")
        operator = data.get("Operator")
        values = data.get("Values")
        return cls.validate(name=name, operator=operator, values=values)  # type: ignore[arg-type]


class DomainsSortCondition:
    def __init__(self, name: DomainFilterField, sort_order: DomainSortOrder):
        self.name: DomainFilterField = name
        self.sort_order: DomainSortOrder = sort_order

    @classmethod
    def validate(cls, name: str, sort_order: str):  # type: ignore[misc,no-untyped-def]
        input_errors: List[str] = []
        if not is_valid_enum(name, DomainFilterField):
            input_errors.append(f"Cannot sort by field {name}")

        if not is_valid_enum(sort_order, DomainSortOrder):
            input_errors.append(f"Invalid sort order {sort_order}")

        if input_errors:
            raise ValidationException(input_errors)

        return cls(name=DomainFilterField(name), sort_order=DomainSortOrder(sort_order))

    @classmethod
    def validate_dict(cls, data: Dict[str, Any]):  # type: ignore[misc,no-untyped-def]
        name = data.get("Name")
        sort_order = data.get("SortOrder")
        return cls.validate(name=name, sort_order=sort_order)  # type: ignore[arg-type]



from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    LambdaResource,
    Resource,
    ServiceResource,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_factory import (
    state_task_service_for,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.state_task import (
    StateTask,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.state_task_lambda import (
    StateTaskLambda,
)


def state_task_for(resource: Resource) -> StateTask:
    if not resource:
        raise ValueError("No Resource declaration in State Task.")
    if isinstance(resource, LambdaResource):
        state = StateTaskLambda()
    elif isinstance(resource, ServiceResource):
        state = state_task_service_for(service_name=resource.service_name)
    else:
        raise NotImplementedError(
            f"Resource of type '{type(resource)}' are not supported: '{resource}'."
        )
    return state



"""amp base URL and path."""

from .responses import PrometheusServiceResponse

url_bases = [
    r"https?://aps\.(.+)\.amazonaws\.com",
]


url_paths = {
    "{0}/workspaces$": PrometheusServiceResponse.dispatch,
    "{0}/workspaces/(?P<workspace_id>[^/]+)$": PrometheusServiceResponse.dispatch,
    "{0}/workspaces/(?P<workspace_id>[^/]+)/alias$": PrometheusServiceResponse.dispatch,
    "{0}/workspaces/(?P<workspace_id>[^/]+)/logging$": PrometheusServiceResponse.dispatch,
    "{0}/workspaces/(?P<workspace_id>[^/]+)/rulegroupsnamespaces$": PrometheusServiceResponse.dispatch,
    "{0}/workspaces/(?P<workspace_id>[^/]+)/rulegroupsnamespaces/(?P<name>[^/]+)$": PrometheusServiceResponse.dispatch,
    "{0}/tags/(?P<resource_arn>[^/]+)$": PrometheusServiceResponse.dispatch,
    "{0}/tags/(?P<arn_prefix>[^/]+)/(?P<workspace_id>[^/]+)$": PrometheusServiceResponse.method_dispatch(
        PrometheusServiceResponse.tags  # type: ignore
    ),
    "{0}/tags/(?P<arn_prefix>[^/]+)/(?P<workspace_id>[^/]+)/(?P<ns_name>[^/]+)$": PrometheusServiceResponse.method_dispatch(
        PrometheusServiceResponse.tags  # type: ignore
    ),
}



import decimal
from typing import Any

from moto.stepfunctions.parser.asl.component.intrinsic.argument.function_argument_list import (
    FunctionArgumentList,
)
from moto.stepfunctions.parser.asl.component.intrinsic.function.statesfunction.states_function import (
    StatesFunction,
)
from moto.stepfunctions.parser.asl.component.intrinsic.functionname.state_function_name_types import (
    StatesFunctionNameType,
)
from moto.stepfunctions.parser.asl.component.intrinsic.functionname.states_function_name import (
    StatesFunctionName,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment


def _round_like_java(f: float) -> int:
    # this behaves a bit weird for boundary values
    # AWS stepfunctions is implemented in Java, so we need to adjust the rounding accordingly
    # python by default rounds half to even
    if f >= 0:
        decimal.getcontext().rounding = decimal.ROUND_HALF_UP
    else:
        decimal.getcontext().rounding = decimal.ROUND_HALF_DOWN
    d = decimal.Decimal(f)
    return round(d, 0)


class MathAdd(StatesFunction):
    # Returns the sum of two numbers.
    #
    # For example:
    # With input
    # {
    #    "value1": 111,
    #    "step": -1
    # }
    #
    # Call
    # "value1.$": "States.MathAdd($.value1, $.step)"
    #
    # Returns
    # {"value1": 110 }

    def __init__(self, arg_list: FunctionArgumentList):
        super().__init__(
            states_name=StatesFunctionName(
                function_type=StatesFunctionNameType.MathAdd
            ),
            arg_list=arg_list,
        )
        if arg_list.size != 2:
            raise ValueError(
                f"Expected 2 arguments for function type '{type(self)}', but got: '{arg_list}'."
            )

    @staticmethod
    def _validate_integer_value(value: Any) -> int:
        if not isinstance(value, (int, float)):
            raise TypeError(f"Expected integer value, but got: '{value}'.")
        # If you specify a non-integer value for one or both the arguments,
        # Step Functions will round it off to the nearest integer.

        if isinstance(value, float):
            result = _round_like_java(value)
            return int(result)

        return value

    def _eval_body(self, env: Environment) -> None:
        self.arg_list.eval(env=env)
        args = env.stack.pop()

        a = self._validate_integer_value(args[0])
        b = self._validate_integer_value(args[1])

        res = a + b
        env.stack.append(res)



from __future__ import annotations

import abc
import json
import threading
from typing import Any, Final, List, Optional

from moto.stepfunctions.parser.api import (
    HistoryEventType,
    MapRunFailedEventDetails,
    MapRunStartedEventDetails,
    MapRunStatus,
)
from moto.stepfunctions.parser.asl.component.common.comment import Comment
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.flow.start_at import StartAt
from moto.stepfunctions.parser.asl.component.program.program import Program
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.item_reader_decl import (
    ItemReader,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.inline_iteration_component import (
    InlineIterationComponent,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.itemprocessor.map_run_record import (
    MapRunRecord,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iteration_worker import (
    IterationWorker,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.job import (
    Job,
    JobPool,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.max_concurrency import (
    MaxConcurrency,
)
from moto.stepfunctions.parser.asl.component.states import States
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.eval.event.event_history import EventHistory


class DistributedIterationComponentEvalInput:
    state_name: Final[str]
    max_concurrency: Final[int]
    item_reader: Final[ItemReader]

    def __init__(self, state_name: str, max_concurrency: int, item_reader: ItemReader):
        self.state_name = state_name
        self.max_concurrency = max_concurrency
        self.item_reader = item_reader


class DistributedIterationComponent(InlineIterationComponent, abc.ABC):
    _eval_input: Optional[DistributedIterationComponentEvalInput]
    _mutex: Final[threading.Lock]
    _map_run_record: Optional[MapRunRecord]
    _workers: List[IterationWorker]

    def __init__(self, start_at: StartAt, states: States, comment: Comment):
        super().__init__(start_at=start_at, states=states, comment=comment)
        self._mutex = threading.Lock()
        self._map_run_record = None
        self._workers = list()

    @abc.abstractmethod
    def _create_worker(self, env: Environment) -> IterationWorker: ...

    def _launch_worker(self, env: Environment) -> IterationWorker:
        worker = super()._launch_worker(env=env)
        self._workers.append(worker)
        return worker

    def _set_active_workers(self, workers_number: int, env: Environment) -> None:
        with self._mutex:
            current_workers_number = len(self._workers)
            workers_diff = workers_number - current_workers_number
            if workers_diff > 0:
                for _ in range(workers_diff):
                    self._launch_worker(env=env)
            elif workers_diff < 0:
                deletion_workers = list(self._workers)[workers_diff:]
                for worker in deletion_workers:
                    worker.sig_stop()
                    self._workers.remove(worker)

    def _map_run(self, env: Environment) -> None:
        input_items: List[json] = env.stack.pop()

        input_item_prog: Final[Program] = Program(
            start_at=self._start_at,
            states=self._states,
            timeout_seconds=None,
            comment=self._comment,
        )
        self._job_pool = JobPool(job_program=input_item_prog, job_inputs=input_items)

        # TODO: add watch on map_run_record update event and adjust the number of running workers accordingly.
        max_concurrency = self._map_run_record.max_concurrency
        workers_number = (
            len(input_items)
            if max_concurrency == MaxConcurrency.DEFAULT
            else max_concurrency
        )
        self._set_active_workers(workers_number=workers_number, env=env)

        self._job_pool.await_jobs()

        worker_exception: Optional[Exception] = self._job_pool.get_worker_exception()
        if worker_exception is not None:
            raise worker_exception

        closed_jobs: List[Job] = self._job_pool.get_closed_jobs()
        outputs: List[Any] = [closed_job.job_output for closed_job in closed_jobs]

        env.stack.append(outputs)

    def _eval_body(self, env: Environment) -> None:
        self._eval_input = env.stack.pop()

        self._map_run_record = MapRunRecord(
            state_machine_arn=env.context_object_manager.context_object["StateMachine"][
                "Id"
            ],
            execution_arn=env.context_object_manager.context_object["Execution"]["Id"],
            max_concurrency=self._eval_input.max_concurrency,
        )
        env.map_run_record_pool_manager.add(self._map_run_record)

        env.event_history.add_event(
            context=env.event_history_context,
            hist_type_event=HistoryEventType.MapRunStarted,
            event_detail=EventDetails(
                mapRunStartedEventDetails=MapRunStartedEventDetails(
                    mapRunArn=self._map_run_record.map_run_arn
                )
            ),
        )

        execution_event_history = env.event_history
        try:
            self._eval_input.item_reader.eval(env=env)
            # TODO: investigate if this is truly propagated also to eventual sub programs in map run states.
            env.event_history = EventHistory()
            self._map_run(env=env)

        except FailureEventException as failure_event_ex:
            map_run_fail_event_detail = MapRunFailedEventDetails()

            maybe_error_cause_pair = failure_event_ex.extract_error_cause_pair()
            if maybe_error_cause_pair:
                error, cause = maybe_error_cause_pair
                if error:
                    map_run_fail_event_detail["error"] = error
                if cause:
                    map_run_fail_event_detail["cause"] = cause

            env.event_history = execution_event_history
            env.event_history.add_event(
                context=env.event_history_context,
                hist_type_event=HistoryEventType.MapRunFailed,
                event_detail=EventDetails(
                    mapRunFailedEventDetails=map_run_fail_event_detail
                ),
            )
            self._map_run_record.set_stop(status=MapRunStatus.FAILED)
            raise failure_event_ex

        except Exception as ex:
            env.event_history = execution_event_history
            env.event_history.add_event(
                context=env.event_history_context,
                hist_type_event=HistoryEventType.MapRunFailed,
                event_detail=EventDetails(
                    mapRunFailedEventDetails=MapRunFailedEventDetails()
                ),
            )
            self._map_run_record.set_stop(status=MapRunStatus.FAILED)
            raise ex
        finally:
            env.event_history = execution_event_history

        # TODO: review workflow of program stops and maprunstops
        # program_state = env.program_state()
        # if isinstance(program_state, ProgramSucceeded)
        env.event_history.add_event(
            context=env.event_history_context,
            hist_type_event=HistoryEventType.MapRunSucceeded,
        )
        self._map_run_record.set_stop(status=MapRunStatus.SUCCEEDED)



from .responses import MediaConnectResponse

url_bases = [
    r"https?://mediaconnect\.(.+)\.amazonaws.com",
]


response = MediaConnectResponse()


url_paths = {
    "{0}/v1/flows$": response.dispatch,
    "{0}/v1/flows/(?P<flowarn>[^/.]+)$": response.dispatch,
    "{0}/v1/flows/(?P<flowarn>[^/.]+)/vpcInterfaces$": response.dispatch,
    "{0}/v1/flows/(?P<flowarn>[^/.]+)/vpcInterfaces/(?P<vpcinterfacename>[^/.]+)$": response.dispatch,
    "{0}/v1/flows/(?P<flowarn>[^/.]+)/source$": response.dispatch,
    "{0}/v1/flows/(?P<flowarn>[^/.]+)/source/(?P<sourcearn>[^/.]+)$": response.dispatch,
    "{0}/v1/flows/(?P<flowarn>[^/.]+)/outputs$": response.dispatch,
    "{0}/v1/flows/(?P<flowarn>[^/.]+)/outputs/(?P<outputarn>[^/.]+)$": response.dispatch,
    "{0}/v1/flows/(?P<flowarn>[^/.]+)/entitlements$": response.dispatch,
    "{0}/v1/flows/(?P<flowarn>[^/.]+)/entitlements/(?P<entitlementarn>[^/.]+)$": response.dispatch,
    "{0}/v1/flows/start/(?P<flowarn>[^/.]+)$": response.dispatch,
    "{0}/v1/flows/stop/(?P<flowarn>[^/.]+)$": response.dispatch,
    "{0}/tags/(?P<resourcearn>[^/.]+)$": response.dispatch,
}



from moto.stepfunctions.parser.asl.component.common.payload.payloadvalue.payloadvaluelit.payload_value_lit import (
    PayloadValueLit,
)


class PayloadValueInt(PayloadValueLit):
    val: int

    def __init__(self, val: int):
        super().__init__(val=val)



from typing import Optional

from antlr4 import ParserRuleContext
from antlr4.tree.Tree import ParseTree, TerminalNodeImpl


class Antlr4Utils:
    @staticmethod
    def is_production(
        pt: ParseTree, rule_index: Optional[int] = None
    ) -> Optional[ParserRuleContext]:
        if isinstance(pt, ParserRuleContext):
            prc = pt.getRuleContext()  # noqa
            if rule_index is not None:
                return prc if prc.getRuleIndex() == rule_index else None
            return prc
        return None

    @staticmethod
    def is_terminal(
        pt: ParseTree, token_type: Optional[int] = None
    ) -> Optional[TerminalNodeImpl]:
        if isinstance(pt, TerminalNodeImpl):
            if token_type is not None:
                return pt if pt.getSymbol().type == token_type else None
            return pt
        return None



import types
from http.client import responses as http_responses
from io import BytesIO
from typing import Any, Dict, List, Optional, Tuple
from urllib.parse import urlparse

import responses
from werkzeug.wrappers import Request

from moto.backends import get_service_from_url
from moto.core.config import passthrough_service, passthrough_url
from moto.core.versions import is_responses_0_17_x

from .responses import TYPE_RESPONSE


class CallbackResponse(responses.CallbackResponse):
    """
    Need to subclass so we can change a couple things
    """

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, CallbackResponse):
            return self.method == other.method and self.url.pattern == other.url.pattern  # type: ignore
        return super().__eq__(other)

    def get_response(self, request: Any) -> responses.HTTPResponse:
        """
        Need to override this so we can pass decode_content=False
        """
        if not isinstance(request, Request):
            url = urlparse(request.url)
            if request.body is None:
                body = None
            elif isinstance(request.body, str):
                body = request.body.encode("UTF-8")
            elif hasattr(request.body, "read"):
                body = request.body.read()
            else:
                body = request.body
            req = Request.from_values(
                path="?".join([url.path, url.query]),
                input_stream=BytesIO(body) if body else None,
                content_length=request.headers.get("Content-Length"),
                content_type=request.headers.get("Content-Type"),
                method=request.method,
                base_url=f"{url.scheme}://{url.netloc}",
                headers=[(k, v) for k, v in request.headers.items()],
            )
            request = req
        headers = self.get_headers()

        from moto.moto_api import recorder

        recorder._record_request(request, body)

        result = self.callback(request)
        if isinstance(result, Exception):
            raise result

        status, r_headers, body = result
        body_io = responses._handle_body(body)
        headers.update(r_headers)

        return responses.HTTPResponse(
            status=status,
            reason=http_responses.get(status),
            body=body_io,
            headers=headers,
            preload_content=False,
            # Need to not decode_content to mimic requests
            decode_content=False,
        )

    def matches(self, request: "responses.PreparedRequest") -> Tuple[bool, str]:
        if request.method != self.method:
            return False, "Method does not match"

        if not self._url_matches(self.url, str(request.url)):
            return False, "URL does not match"

        service = get_service_from_url(request.url)  # type: ignore
        if (service and passthrough_service(service)) or passthrough_url(request.url):  # type: ignore
            return False, "URL does not match"

        return super().matches(request)

    def _url_matches(
        self, url: Any, other: Any, match_querystring: bool = False
    ) -> bool:
        """
        Need to override this so we can fix querystrings breaking regex matching
        """
        if not match_querystring:
            other = other.split("?", 1)[0]

        if isinstance(url, str):
            if responses._has_unicode(url):
                url = responses._clean_unicode(url)
                if not isinstance(other, str):
                    other = other.encode("ascii").decode("utf8")
            return self._url_matches_strict(url, other)  # type: ignore[attr-defined]
        elif isinstance(url, responses.Pattern) and url.match(other):
            return True
        else:
            return False


def not_implemented_callback(
    request: Any,  # pylint: disable=unused-argument
) -> TYPE_RESPONSE:
    status = 400
    headers: Dict[str, str] = {}
    response = "The method is not implemented"

    return status, headers, response


# Modify behaviour of the matcher to only/always return the first match
# Default behaviour is to return subsequent matches for subsequent requests, which leads to https://github.com/getmoto/moto/issues/2567
#  - First request matches on the appropriate S3 URL
#  - Same request, executed again, will be matched on the subsequent match, which happens to be the catch-all, not-yet-implemented, callback
# Fix: Always return the first match
#
# Note that, due to an outdated API we no longer support Responses <= 0.12.1
# This method should be used for Responses 0.12.1 < .. < 0.17.0
def _find_first_match(
    self: Any, request: Any
) -> Tuple[Optional[responses.BaseResponse], List[str]]:
    matches = []
    match_failed_reasons = []
    all_possibles = self._matches + responses._default_mock._matches  # type: ignore[attr-defined]
    for match in all_possibles:
        match_result, reason = match.matches(request)
        if match_result:
            matches.append(match)
        else:
            match_failed_reasons.append(reason)

    # Look for implemented callbacks first
    implemented_matches = [
        m
        for m in matches
        if type(m) is not CallbackResponse or m.callback != not_implemented_callback
    ]
    if implemented_matches:
        return implemented_matches[0], []
    elif matches:
        # We had matches, but all were of type not_implemented_callback
        return matches[0], match_failed_reasons

    return None, match_failed_reasons


def get_response_mock() -> responses.RequestsMock:
    """
    The responses-library is crucial in ensuring that requests to AWS are intercepted, and routed to the right backend.
    However, as our usecase is different from a 'typical' responses-user, Moto always needs some custom logic to ensure responses behaves in a way that works for us.

    For older versions, that meant changing the internal logic
    For later versions, > 0.17.0, we can use a custom registry, and extend the logic instead of overriding it

    For all versions, we need to add passthrough to allow non-AWS requests to work
    """
    responses_mock = None

    if is_responses_0_17_x():
        from .responses_custom_registry import CustomRegistry

        responses_mock = responses.RequestsMock(
            assert_all_requests_are_fired=False, registry=CustomRegistry
        )
    else:
        responses_mock = responses.RequestsMock(assert_all_requests_are_fired=False)
        responses_mock._find_match = types.MethodType(_find_first_match, responses_mock)  # type: ignore[method-assign]

    responses_mock.add_passthru("http")
    return responses_mock


def reset_responses_mock(responses_mock: responses.RequestsMock) -> None:
    if is_responses_0_17_x():
        from .responses_custom_registry import CustomRegistry

        responses_mock.reset()
        # No way to set the registry directly (yet..)
        responses_mock._set_registry(CustomRegistry)
        responses_mock.add_passthru("http")
    else:
        responses_mock.reset()



"""WorkSpacesBackend class with methods for supported APIs."""

import re
from collections.abc import Mapping
from typing import Any, Dict, List, Optional, Tuple

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.core.utils import unix_time
from moto.ds import ds_backends
from moto.ds.models import Directory
from moto.ec2 import ec2_backends
from moto.moto_api._internal import mock_random
from moto.workspaces.exceptions import (
    InvalidParameterValuesException,
    ResourceAlreadyExistsException,
    ResourceNotFoundException,
    ValidationException,
)


class Workspace(BaseModel):
    def __init__(
        self,
        workspace: Dict[str, Any],
        running_mode: str,
        error_code: str,
        error_msg: str,
    ):
        self.workspace_properties: Dict[str, Any]
        self.workspace = workspace
        self.workspace_id = f"ws-{mock_random.get_random_hex(9)}"
        # Create_workspaces operation is asynchronous and returns before the WorkSpaces are created.
        # Initially the 'state' is 'PENDING', but here the 'state' will be set as 'AVAILABLE' since
        # this operation is being mocked.
        self.directory_id = workspace["DirectoryId"]
        self.bundle_id = workspace["BundleId"]
        self.user_name = workspace["UserName"]
        self.state = "AVAILABLE"
        self.error_message = error_msg or ""
        self.error_code = error_code or ""
        self.volume_encryption_key = workspace.get("VolumeEncryptionKey", "")
        self.user_volume_encryption_enabled = workspace.get(
            "UserVolumeEncryptionEnabled", ""
        )
        self.root_volume_encryption_enabled = workspace.get(
            "RootVolumeEncryptionEnabled", ""
        )
        workspace_properties = {"RunningMode": running_mode}
        self.workspace_properties = workspace.get("WorkspaceProperties", "")

        if self.workspace_properties:
            self.workspace_properties["RunningMode"] = running_mode
        else:
            self.workspace_properties = workspace_properties

        self.computer_name = ""  # Workspace Bundle
        self.modification_states: List[
            Dict[str, str]
        ] = []  # modify_workspace_properties
        self.related_workspaces: List[Dict[str, str]] = []  # create_standy_workspace
        self.data_replication_settings: Dict[str, Any] = {}
        # The properties of the standby WorkSpace related to related_workspaces
        self.standby_workspaces_properties: List[Dict[str, Any]] = []
        self.tags = workspace.get("Tags", [])

    def to_dict_pending(self) -> Dict[str, Any]:
        dct = {
            "WorkspaceId": self.workspace_id,
            "DirectoryId": self.directory_id,
            "UserName": self.user_name,
            "IpAddress": "",  # UnKnown
            "State": self.state,
            "BundleId": self.bundle_id,
            "SubnetId": "",  # UnKnown
            "ErrorMessage": self.error_message,
            "ErrorCode": self.error_code,
            "ComputerName": self.computer_name,
            "VolumeEncryptionKey": self.volume_encryption_key,
            "UserVolumeEncryptionEnabled": self.user_volume_encryption_enabled,
            "RootVolumeEncryptionEnabled": self.root_volume_encryption_enabled,
            "WorkspaceProperties": self.workspace_properties,
            "ModificationStates": self.modification_states,
            "RelatedWorkspaces": self.related_workspaces,
            "DataReplicationSettings": self.data_replication_settings,
            "StandbyWorkspacesProperties": self.standby_workspaces_properties,
        }
        return {k: v for k, v in dct.items() if v}

    def filter_empty_values(self, d: Dict[str, Any]) -> Dict[str, Any]:
        if isinstance(d, Mapping):
            return dict((k, self.filter_empty_values(v)) for k, v in d.items() if v)
        else:
            return d

    def to_dict_failed(self) -> Dict[str, Any]:
        dct = {
            "WorkspaceRequest": {
                "DirectoryId": self.workspace["DirectoryId"],
                "UserName": self.workspace["UserName"],
                "BundleId": self.workspace["BundleId"],
                "SubnetId": "",  # UnKnown
                "VolumeEncryptionKey": self.volume_encryption_key,
                "UserVolumeEncryptionEnabled": self.user_volume_encryption_enabled,
                "RootVolumeEncryptionEnabled": self.root_volume_encryption_enabled,
                "WorkspaceProperties": self.workspace_properties,
                "Tags": self.tags,
            },
            "ErrorCode": self.error_code,
            "ErrorMessage": self.error_message,
        }
        return self.filter_empty_values(dct)


class WorkSpaceDirectory(BaseModel):
    def __init__(
        self,
        account_id: str,
        region: str,
        directory: Directory,
        registration_code: str,
        security_group_id: str,
        subnet_ids: List[str],
        enable_work_docs: bool,
        enable_self_service: bool,
        tenancy: str,
        tags: List[Dict[str, str]],
    ):
        self.account_id = account_id
        self.region = region
        self.directory_id = directory.directory_id
        self.alias = directory.alias
        self.directory_name = directory.name
        self.launch_time = directory.launch_time
        self.registration_code = registration_code
        if directory.directory_type == "ADConnector":
            dir_subnet_ids = directory.connect_settings["SubnetIds"]  # type: ignore[index]
        else:
            dir_subnet_ids = directory.vpc_settings["SubnetIds"]  # type: ignore[index]
        self.subnet_ids = subnet_ids or dir_subnet_ids
        self.dns_ip_addresses = directory.dns_ip_addrs
        self.customer_username = "Administrator"
        self.iam_rold_id = f"arn:aws:iam::{account_id}:role/workspaces_DefaultRole"
        self.directory_type = directory.directory_type
        self.workspace_security_group_id = security_group_id
        self.state = "REGISTERED"
        # Default values for workspace_creation_properties
        workspace_creation_properties = {
            "EnableWorkDocs": enable_work_docs,
            "EnableInternetAccess": False,
            "DefaultOu": "",
            "CustomSecurityGroupId": "",
            "UserEnabledAsLocalAdministrator": (
                True if self.customer_username == "Administrator" else False
            ),
            "EnableMaintenanceMode": True,
        }
        # modify creation properites
        self.workspace_creation_properties = workspace_creation_properties
        self.ip_group_ids = ""  # create_ip_group
        # Default values for workspace access properties
        workspace_access_properties = {
            "DeviceTypeWindows": (
                "DENY" if self.directory_type == "AD_CONNECTOR" else "ALLOW"
            ),
            "DeviceTypeOsx": "ALLOW",
            "DeviceTypeWeb": "DENY",
            "DeviceTypeIos": "ALLOW",
            "DeviceTypeAndroid": "ALLOW",
            "DeviceTypeChromeOs": "ALLOW",
            "DeviceTypeZeroClient": (
                "DENY" if self.directory_type == "AD_CONNECTOR" else "ALLOW"
            ),
            "DeviceTypeLinux": "DENY",
        }
        # modify_workspace_access_properties
        self.workspace_access_properties = workspace_access_properties
        self.tenancy = tenancy or "SHARED"

        # Default values for self service permissions
        mode = "DISABLED"
        if enable_self_service:
            mode = "ENABLED"
        self_service_permissions = {
            "RestartWorkspace": "ENABLED",
            "IncreaseVolumeSize": mode,
            "ChangeComputeType": mode,
            "SwitchRunningMode": mode,
            "RebuildWorkspace": mode,
        }
        self.self_service_permissions = self_service_permissions
        # Default values for saml properties
        saml_properties = {
            "Status": "DISABLED",
            "UserAccessUrl": "",
            "RelayStateParameterName": "RelayState",
        }
        self.saml_properties = saml_properties
        # Default values for certificate bases auth properties
        self.certificate_based_auth_properties = {
            "Status": "DISABLED",
        }
        # ModifyCertificateBasedAuthProperties
        self.tags = tags or []
        client_properties = {
            # Log uploading is enabled by default.
            "ReconnectEnabled": "ENABLED",
            "LogUploadEnabled": "ENABLED",  # Remember me is enabled by default
        }
        self.client_properties = client_properties

    def delete_security_group(self) -> None:
        """Delete the given security group."""
        ec2_backends[self.account_id][self.region].delete_security_group(
            group_id=self.workspace_security_group_id
        )

    def to_dict(self) -> Dict[str, Any]:
        dct = {
            "DirectoryId": self.directory_id,
            "Alias": self.alias,
            "DirectoryName": self.directory_name,
            "RegistrationCode": self.registration_code,
            "SubnetIds": self.subnet_ids,
            "DnsIpAddresses": self.dns_ip_addresses,
            "CustomerUserName": self.customer_username,
            "IamRoleId": self.iam_rold_id,
            "DirectoryType": self.directory_type,
            "WorkspaceSecurityGroupId": self.workspace_security_group_id,
            "State": self.state,
            "WorkspaceCreationProperties": self.workspace_creation_properties,
            "ipGroupIds": self.ip_group_ids,
            "WorkspaceAccessProperties": self.workspace_access_properties,
            "Tenancy": self.tenancy,
            "SelfservicePermissions": self.self_service_permissions,
            "SamlProperties": self.saml_properties,
            "CertificateBasedAuthProperties": self.certificate_based_auth_properties,
        }
        return {k: v for k, v in dct.items() if v}


class WorkspaceImage(BaseModel):
    def __init__(
        self,
        name: str,
        description: str,
        tags: List[Dict[str, str]],
        account_id: str,
    ):
        self.image_id = f"wsi-{mock_random.get_random_hex(9)}"
        self.name = name
        self.description = description
        self.operating_system: Dict[str, str] = {}  # Unknown
        # Initially the 'state' is 'PENDING', but here the 'state' will be set as 'AVAILABLE' since
        # this operation is being mocked.
        self.state = "AVAILABLE"
        self.required_tenancy = "DEFAULT"
        self.created = unix_time()
        self.owner_account = account_id
        self.error_code = ""
        self.error_message = ""
        self.image_permissions: List[Dict[str, str]] = []
        self.tags = tags

        # Default updates
        self.updates = {
            "UpdateAvailable": False,
            "Description": "This WorkSpace image does not have updates available",
        }
        self.error_details: List[Dict[str, str]] = []

    def to_dict(self) -> Dict[str, Any]:
        dct = {
            "ImageId": self.image_id,
            "Name": self.name,
            "Description": self.description,
            "OperatingSystem": self.operating_system,
            "State": self.state,
            "RequiredTenancy": self.required_tenancy,
            "Created": self.created,
            "OwnerAccountId": self.owner_account,
        }
        return {k: v for k, v in dct.items() if v}

    def to_desc_dict(self) -> Dict[str, Any]:
        dct = self.to_dict()
        dct_options = {
            "ErrorCode": self.error_code,
            "ErrorMessage": self.error_message,
            "Updates": self.updates,
            "ErrorDetails": self.error_details,
        }
        for key, value in dct_options.items():
            if value is not None:
                dct[key] = value
        return dct


class WorkSpacesBackend(BaseBackend):
    """Implementation of WorkSpaces APIs."""

    # The assumption here is that the limits are the same for all regions.
    DIRECTORIES_LIMIT = 50

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.workspaces: Dict[str, Workspace] = dict()
        self.workspace_directories: Dict[str, WorkSpaceDirectory] = dict()
        self.workspace_images: Dict[str, WorkspaceImage] = dict()
        self.directories: List[Directory]

    def validate_directory_id(self, value: str, msg: str) -> None:
        """Raise exception if the directory id is invalid."""
        id_pattern = r"^d-[0-9a-f]{10}$"
        if not re.match(id_pattern, value):
            raise ValidationException(msg)

    def validate_image_id(self, value: str, msg: str) -> None:
        """Raise exception if the image id is invalid."""
        id_pattern = r"^wsi-[0-9a-z]{9}$"
        if not re.match(id_pattern, value):
            raise ValidationException(msg)

    def create_security_group(self, directory_id: str, vpc_id: str) -> str:
        """Create security group for the workspace directory."""
        security_group_info = ec2_backends[self.account_id][
            self.region_name
        ].create_security_group(
            name=f"{directory_id}_workspacesMembers",
            description=("Amazon WorkSpaces Security Group"),
            vpc_id=vpc_id,
        )
        return security_group_info.id

    def create_workspaces(
        self, workspaces: List[Dict[str, Any]]
    ) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
        failed_requests = []
        pending_requests = []

        for ws in workspaces:
            error_code = ""
            error_msg = ""
            directory_id = ws["DirectoryId"]
            msg = f"The Directory ID {directory_id} in the request is invalid."
            self.validate_directory_id(directory_id, msg)

            # FailedRequests are created if the directory_id is unknown
            if directory_id not in self.workspace_directories:
                error_code = "ResourceNotFound.Directory"
                error_msg = "The specified directory could not be found in the specified region."

            running_mode = "ALWAYS_ON"
            workspace_properties = ws.get("WorkspaceProperties", "")
            if workspace_properties:
                running_mode = workspace_properties.get("RunningMode", running_mode)
                auto_stop_timeout = workspace_properties.get(
                    "RunningModeAutoStopTimeoutInMinutes", ""
                )

                # Requests fail if AutoStopTimeout is given for an AlwaysOn Running mode
                if auto_stop_timeout and running_mode == "ALWAYS_ON":
                    error_code = "AutoStopTimeoutIsNotApplicableForAnAlwaysOnWorkspace"
                    error_msg = "RunningModeAutoStopTimeoutInMinutes is not applicable for WorkSpace with running mode set to ALWAYS_ON."

                # Requests fail if AutoStopTimeout is given for an Manual Running mode
                if auto_stop_timeout and running_mode == "MANUAL":
                    error_code = "AutoStopTimeoutIsNotDefaultForManualWorkspace"

            workspace = Workspace(
                workspace=ws,
                running_mode=running_mode,
                error_code=error_code,
                error_msg=error_msg,
            )
            if error_code:
                failed_requests.append(workspace.to_dict_failed())
            else:
                pending_requests.append(workspace.to_dict_pending())
                self.workspaces[workspace.workspace_id] = workspace

        return failed_requests, pending_requests

    def describe_workspaces(
        self,
        workspace_ids: List[str],
        directory_id: str,
        user_name: str,
        bundle_id: str,
    ) -> List[Workspace]:
        # Pagination not yet implemented

        # Only one of the following are allowed to be specified: BundleId, DirectoryId, WorkSpaceIds.
        if (
            (workspace_ids and directory_id)
            or (directory_id and bundle_id)
            or (workspace_ids and bundle_id)
        ):
            msg = "An invalid number of parameters provided with DescribeWorkspaces. Only one of the following are allowed to be specified: BundleId, DirectoryId, WorkSpaceIds, Filters."
            raise InvalidParameterValuesException(msg)

        # Directory_id parameter is required when Username is given.
        if user_name and not directory_id:
            msg = "The DirectoryId parameter is required when UserName is used."
            raise InvalidParameterValuesException(msg)

        workspaces = list(self.workspaces.values())
        if workspace_ids:
            workspaces = [x for x in workspaces if x.workspace_id in workspace_ids]
        if directory_id:
            workspaces = [x for x in workspaces if x.directory_id == directory_id]
        if directory_id and user_name:
            workspaces = [
                x
                for x in workspaces
                if (x.directory_id == directory_id) and (x.user_name == user_name)
            ]
        if bundle_id:
            workspaces = [x for x in workspaces if x.bundle_id == bundle_id]
        # workspaces = [w.to_dict_pending() for w in workspaces]
        return workspaces

    def register_workspace_directory(
        self,
        directory_id: str,
        subnet_ids: List[str],
        enable_work_docs: bool,
        enable_self_service: bool,
        tenancy: str,
        tags: List[Dict[str, str]],
    ) -> None:
        ran_str = mock_random.get_random_string(length=6)
        registration_code = f"SLiad+{ran_str.upper()}"

        (self.directories, _) = ds_backends[self.account_id][
            self.region_name
        ].describe_directories(directory_ids=[directory_id])
        directory = self.directories[0]

        if directory.directory_type == "ADConnector":
            vpc_id = directory.connect_settings["VpcId"]  # type: ignore[index]
        else:
            vpc_id = directory.vpc_settings["VpcId"]  # type: ignore[index]

        security_group_id = self.create_security_group(directory_id, vpc_id)

        workspace_directory = WorkSpaceDirectory(
            account_id=self.account_id,
            region=self.region_name,
            directory=directory,
            registration_code=registration_code,
            security_group_id=security_group_id,
            subnet_ids=subnet_ids,
            enable_work_docs=enable_work_docs,
            enable_self_service=enable_self_service,
            tenancy=tenancy,
            tags=tags,
        )
        self.workspace_directories[workspace_directory.directory_id] = (
            workspace_directory
        )

    def describe_workspace_directories(
        self, directory_ids: Optional[List[str]] = None
    ) -> List[WorkSpaceDirectory]:
        """Return info on all directories or directories with matching IDs."""
        # Pagination not yet implemented

        workspace_directories = list(self.workspace_directories.values())
        if directory_ids:
            for d in directory_ids:
                msg = "The request is invalid."
                self.validate_directory_id(d, msg)
            workspace_directories = [
                x for x in workspace_directories if x.directory_id in directory_ids
            ]

        return sorted(workspace_directories, key=lambda x: x.launch_time)

    def modify_workspace_creation_properties(
        self, resource_id: str, workspace_creation_properties: Dict[str, Any]
    ) -> None:
        # Raise Exception if Directory doesnot exist.
        if resource_id not in self.workspace_directories:
            raise ValidationException("The request is invalid.")

        res = self.workspace_directories[resource_id]
        res.workspace_creation_properties = workspace_creation_properties

    def create_tags(self, resource_id: str, tags: List[Dict[str, str]]) -> None:
        if resource_id.startswith("d-"):
            ds = self.workspace_directories[resource_id]
            ds.tags.extend(tags)
        if resource_id.startswith("ws-"):
            ws = self.workspaces[resource_id]
            ws.tags.extend(tags)

    def describe_tags(self, resource_id: str) -> List[Dict[str, str]]:
        if resource_id.startswith("d-"):
            ds = self.workspace_directories[resource_id]
            tag_list = ds.tags
        if resource_id.startswith("ws-"):
            ws = self.workspaces[resource_id]
            tag_list = ws.tags
        if resource_id.startswith("wsi-"):
            wsi = self.workspace_images[resource_id]
            tag_list = wsi.tags
        return tag_list

    def describe_client_properties(self, resource_ids: str) -> List[Dict[str, Any]]:
        workspace_directories = list(self.workspace_directories.values())
        workspace_directories = [
            x for x in workspace_directories if x.directory_id in resource_ids
        ]
        client_properties_list = []
        for wd in workspace_directories:
            cpl = {
                "ResourceId": wd.directory_id,
                "ClientProperties": wd.client_properties,
            }
            client_properties_list.append(cpl)
        return client_properties_list

    def modify_client_properties(
        self, resource_id: str, client_properties: Dict[str, str]
    ) -> None:
        res = self.workspace_directories[resource_id]
        res.client_properties = client_properties

    def create_workspace_image(
        self, name: str, description: str, workspace_id: str, tags: List[Dict[str, str]]
    ) -> Dict[str, Any]:
        # Check if workspace exists.
        if workspace_id not in self.workspaces:
            raise ResourceNotFoundException(
                "The specified WorkSpace cannot be found. Confirm that the workspace exists in your AWS account, and try again."
            )
        # Check if image name already exists.
        if name in [x.name for x in self.workspace_images.values()]:
            raise ResourceAlreadyExistsException(
                "A WorkSpace image with the same name exists in the destination Region. Provide a unique destination image name, and try again."
            )

        workspace_image = WorkspaceImage(
            name=name,
            description=description,
            tags=tags,
            account_id=self.account_id,
        )
        self.workspace_images[workspace_image.image_id] = workspace_image
        return workspace_image.to_dict()

    def describe_workspace_images(
        self, image_ids: Optional[List[str]], image_type: Optional[str]
    ) -> List[Dict[str, Any]]:
        # Pagination not yet implemented
        workspace_images = list(self.workspace_images.values())
        if image_type == "OWNED":
            workspace_images = [
                i for i in workspace_images if i.owner_account == self.account_id
            ]
        elif image_type == "SHARED":
            workspace_images = [
                i for i in workspace_images if i.owner_account != self.account_id
            ]
        if image_ids:
            workspace_images = [i for i in workspace_images if i.image_id in image_ids]
        return [w.to_desc_dict() for w in workspace_images]

    def update_workspace_image_permission(
        self, image_id: str, allow_copy_image: bool, shared_account_id: str
    ) -> None:
        shared_account = {"SharedAccountId": shared_account_id}
        res = self.workspace_images[image_id]
        shared_accounts = []
        shared_accounts = res.image_permissions

        if shared_account not in shared_accounts and allow_copy_image:
            shared_accounts.append(shared_account)
        if shared_account in shared_accounts and not allow_copy_image:
            shared_accounts.remove(shared_account)

        res.image_permissions = shared_accounts

    def describe_workspace_image_permissions(
        self, image_id: str
    ) -> Tuple[str, List[Dict[str, str]]]:
        # Pagination not yet implemented

        msg = f"The Image ID {image_id} in the request is invalid"
        self.validate_image_id(image_id, msg)

        image_permissions = []
        if image_id in self.workspace_images:
            res = self.workspace_images[image_id]
            image_permissions = res.image_permissions
        return image_id, image_permissions

    def deregister_workspace_directory(self, directory_id: str) -> None:
        """Deregister Workspace Directory with the matching ID."""
        # self._validate_directory_id(directory_id)
        self.workspace_directories[directory_id].delete_security_group()
        self.workspace_directories.pop(directory_id)

    def modify_selfservice_permissions(
        self, resource_id: str, selfservice_permissions: Dict[str, str]
    ) -> None:
        res = self.workspace_directories[resource_id]
        res.self_service_permissions = selfservice_permissions


workspaces_backends = BackendDict(WorkSpacesBackend, "workspaces")



import hashlib
from typing import Final

from moto.stepfunctions.parser.asl.component.intrinsic.argument.function_argument_list import (
    FunctionArgumentList,
)
from moto.stepfunctions.parser.asl.component.intrinsic.function.statesfunction.hash_calculations.hash_algorithm import (
    HashAlgorithm,
)
from moto.stepfunctions.parser.asl.component.intrinsic.function.statesfunction.states_function import (
    StatesFunction,
)
from moto.stepfunctions.parser.asl.component.intrinsic.functionname.state_function_name_types import (
    StatesFunctionNameType,
)
from moto.stepfunctions.parser.asl.component.intrinsic.functionname.states_function_name import (
    StatesFunctionName,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment


class HashFunc(StatesFunction):
    MAX_INPUT_CHAR_LEN: Final[int] = 10_000

    def __init__(self, arg_list: FunctionArgumentList):
        super().__init__(
            states_name=StatesFunctionName(function_type=StatesFunctionNameType.Hash),
            arg_list=arg_list,
        )
        if arg_list.size != 2:
            raise ValueError(
                f"Expected 2 arguments for function type '{type(self)}', but got: '{arg_list}'."
            )

    @staticmethod
    def _hash_inp_with_alg(inp: str, alg: HashAlgorithm) -> str:
        inp_enc = inp.encode()
        hash_inp = None
        if alg == HashAlgorithm.MD5:
            hash_inp = hashlib.md5(inp_enc)
        if alg == HashAlgorithm.SHA_1:
            hash_inp = hashlib.sha1(inp_enc)
        elif alg == HashAlgorithm.SHA_256:
            hash_inp = hashlib.sha256(inp_enc)
        elif alg == HashAlgorithm.SHA_384:
            hash_inp = hashlib.sha384(inp_enc)
        elif alg == HashAlgorithm.SHA_512:
            hash_inp = hashlib.sha512(inp_enc)
        hash_value: str = hash_inp.hexdigest()
        return hash_value

    def _eval_body(self, env: Environment) -> None:
        self.arg_list.eval(env=env)
        args = env.stack.pop()

        algorithm = args.pop()
        try:
            hash_algorithm = HashAlgorithm(algorithm)
        except Exception:
            raise ValueError(f"Unknown hash function '{algorithm}'.")

        input_data = args.pop()
        if not isinstance(input_data, str):
            raise TypeError(
                f"Expected string type as input data for function type '{type(self)}', but got: '{input_data}'."
            )

        if len(input_data) > self.MAX_INPUT_CHAR_LEN:
            raise ValueError(
                f"Maximum character input length for  for function type '{type(self)}' "
                f"is '{self.MAX_INPUT_CHAR_LEN}', but got '{len(input_data)}'."
            )

        res = self._hash_inp_with_alg(input_data, hash_algorithm)
        env.stack.append(res)



from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.eval.resource_output_transformer.transformer import (
    ResourceOutputTransformer,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.eval.resource_output_transformer.transformer_csv import (
    ResourceOutputTransformerCSV,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.eval.resource_output_transformer.transformer_json import (
    ResourceOutputTransformerJson,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.reader_config.input_type import (
    InputType,
    InputTypeValue,
)


def resource_output_transformer_for(input_type: InputType) -> ResourceOutputTransformer:
    if input_type.input_type_value == InputTypeValue.CSV:
        return ResourceOutputTransformerCSV()
    elif input_type.input_type_value == InputTypeValue.JSON:
        return ResourceOutputTransformerJson()
    else:
        raise ValueError(f"Unknown InputType value: '{input_type.input_type_value}'.")



import copy
import datetime
import threading
from typing import Final, List, Optional

from moto.stepfunctions.parser.api import ExecutionFailedEventDetails, HistoryEventType
from moto.stepfunctions.parser.asl.component.common.error_name.custom_error_name import (
    CustomErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.component.program.program import Program
from moto.stepfunctions.parser.asl.component.state.exec.state_parallel.branch_worker import (
    BranchWorker,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.eval.program_state import ProgramError, ProgramState


class BranchWorkerPool(BranchWorker.BranchWorkerComm):
    _mutex: Final[threading.Lock]
    _termination_event: Final[threading.Event]
    _active_workers_num: int

    _terminated_with_error: Optional[ExecutionFailedEventDetails]

    def __init__(self, workers_num: int):
        self._mutex = threading.Lock()
        self._termination_event = threading.Event()
        self._active_workers_num = workers_num

        self._terminated_with_error = None

    def on_terminated(self, env: Environment):
        if self._termination_event.is_set():
            return
        with self._mutex:
            end_program_state: ProgramState = env.program_state()
            if isinstance(end_program_state, ProgramError):
                self._terminated_with_error = end_program_state.error or dict()
                self._termination_event.set()
            else:
                self._active_workers_num -= 1
                if self._active_workers_num == 0:
                    self._termination_event.set()

    def wait(self):
        self._termination_event.wait()

    def get_exit_event_details(self) -> Optional[ExecutionFailedEventDetails]:
        return self._terminated_with_error


class BranchesDecl(EvalComponent):
    def __init__(self, programs: List[Program]):
        self.programs: Final[List[Program]] = programs

    def _eval_body(self, env: Environment) -> None:
        # Input value for every state_parallel process.
        input_val = env.stack.pop()

        branch_worker_pool = BranchWorkerPool(workers_num=len(self.programs))

        branch_workers: List[BranchWorker] = list()
        for program in self.programs:
            # Environment frame for this sub process.
            env_frame: Environment = env.open_frame()
            env_frame.inp = copy.deepcopy(input_val)

            # Launch the worker.
            worker = BranchWorker(
                branch_worker_comm=branch_worker_pool, program=program, env=env_frame
            )
            branch_workers.append(worker)

            worker.start()

        branch_worker_pool.wait()

        # Propagate exception if parallel task failed.
        exit_event_details: Optional[ExecutionFailedEventDetails] = (
            branch_worker_pool.get_exit_event_details()
        )
        if exit_event_details is not None:
            for branch_worker in branch_workers:
                branch_worker.stop(
                    stop_date=datetime.datetime.now(), cause=None, error=None
                )
                env.close_frame(branch_worker.env)

            exit_error_name = exit_event_details.get("error")
            raise FailureEventException(
                failure_event=FailureEvent(
                    error_name=CustomErrorName(error_name=exit_error_name),
                    event_type=HistoryEventType.ExecutionFailed,
                    event_details=EventDetails(
                        executionFailedEventDetails=exit_event_details
                    ),
                )
            )

        # Collect the results and return.
        result_list = list()

        for worker in branch_workers:
            env_frame = worker.env
            result_list.append(env_frame.inp)
            env.close_frame(env_frame)

        env.stack.append(result_list)



import json
from typing import Dict, Tuple

from moto.core.responses import BaseResponse

from .models import MediaStoreDataBackend, mediastoredata_backends


class MediaStoreDataResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="mediastore-data")

    @property
    def mediastoredata_backend(self) -> MediaStoreDataBackend:
        return mediastoredata_backends[self.current_account][self.region]

    def get_object(self) -> Tuple[str, Dict[str, str]]:
        path = self._get_param("Path")
        result = self.mediastoredata_backend.get_object(path=path)
        headers = {"Path": result.path}
        return result.body, headers

    def put_object(self) -> str:
        body = self.body
        path = self._get_param("Path")
        new_object = self.mediastoredata_backend.put_object(body, path)
        object_dict = new_object.to_dict()
        return json.dumps(object_dict)

    def delete_object(self) -> str:
        item_id = self._get_param("Path")
        self.mediastoredata_backend.delete_object(path=item_id)
        return "{}"

    def list_items(self) -> str:
        items = self.mediastoredata_backend.list_items()
        return json.dumps(dict(Items=items))



import abc
from typing import Any

from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.utils import SubtypesInstanceManager


class Operator(abc.ABC, SubtypesInstanceManager):
    @staticmethod
    @abc.abstractmethod
    def eval(env: Environment, value: Any) -> None:
        pass
